"""
Plugins infrastructure

See README.rst for details.
"""

from .constants import PluginContexts, PluginSettings, PluginSignals, PluginURLs
from .pluggable_override import pluggable_override
from .plugin_apps import get_plugin_apps
from .plugin_contexts import get_plugins_view_context
from .plugin_manager import PluginError, PluginManager
from .plugin_settings import add_plugins
from .plugin_signals import connect_plugin_receivers
from .plugin_urls import get_plugin_url_patterns
from .registry import get_plugin_app_configs



"""Test user functions"""

import pytest
from django.test import TestCase

from edx_django_utils.user import generate_password


class GeneratePasswordTest(TestCase):
    """Tests formation of randomly generated passwords."""

    def test_default_args(self):
        password = generate_password()
        assert 12 == len(password)
        assert any(c.isdigit for c in password)
        assert any(c.isalpha for c in password)

    def test_length(self):
        length = 25
        assert length == len(generate_password(length=length))

    def test_chars(self):
        char = '!'
        password = generate_password(length=12, chars=(char,))

        assert any(c.isdigit for c in password)
        assert any(c.isalpha for c in password)
        assert (char * 10) == password[2:]

    def test_min_length(self):
        with pytest.raises(ValueError):
            generate_password(length=7)



"""
Various functions to get view contexts

Please remember to expose any new public methods in the `__init__.py` file.
"""
import functools
from importlib import import_module
from logging import getLogger

from . import constants, registry

log = getLogger(__name__)


def get_plugins_view_context(project_type, view_name, existing_context=None):
    """
    Returns a dict of additional view context. Will check if any plugin apps
    have that view in their view_context_config, and if so will call their
    selected function to get their context dicts.

    Params:
        project_type: a string that determines which project (lms or studio) the view is being called in. See the
            ProjectType enum in plugins/constants.py for valid options
        view_name: a string that determines which view needs the additional context. These are globally unique and
            noted in the api.py in the view's app.
        existing_context: a dictionary which includes all of the data that the page was going to render with prior
            to the addition of each plugin's context. This is what will be passed to plugins so they may choose
            what data to add to the view.
    """
    aggregate_context = {"plugins": {}}

    if existing_context is None:
        existing_context = {}

    context_functions = _get_cached_context_functions_for_view(project_type, view_name)

    for (context_function, plugin_name) in context_functions:
        try:
            plugin_context = context_function(existing_context)
        except Exception as exc:  # pylint: disable=broad-except
            # We're catching this because we don't want the core to blow up when a
            # plugin is broken. This exception will probably need some sort of
            # monitoring hooked up to it to make sure that these errors don't go
            # unseen.
            log.exception("Failed to call plugin context function. Error: %s", exc)
            continue

        aggregate_context["plugins"][plugin_name] = plugin_context

    return aggregate_context


@functools.lru_cache(maxsize=None)
def _get_cached_context_functions_for_view(project_type, view_name):
    """
    Returns a list of tuples where the first item is the context function
    and the second item is the name of the plugin it's being called from.

    NOTE: These will be functions will be cached (in RAM not memcache) on this unique
    combination. If we enable many new views to use this system, we may notice an
    increase in memory usage as the entirety of these functions will be held in memory.
    """
    context_functions = []
    for app_config in registry.get_plugin_app_configs(project_type):
        context_function_path = _get_context_function_path(
            app_config, project_type, view_name
        )
        if context_function_path:
            module_path, _, name = context_function_path.rpartition(".")
            try:
                module = import_module(module_path)
            except ImportError:
                log.exception(
                    "Failed to import %s plugin when creating %s context",
                    module_path,
                    view_name,
                )
                continue
            context_function = getattr(module, name, None)
            if context_function:
                plugin_name, _, _ = module_path.partition(".")
                context_functions.append((context_function, plugin_name))
            else:
                log.warning(
                    "Failed to retrieve %s function from %s plugin when creating %s context",
                    name,
                    module_path,
                    view_name,
                )
    return context_functions


def _get_context_function_path(app_config, project_type, view_name):
    plugin_config = getattr(app_config, constants.PLUGIN_APP_CLASS_ATTRIBUTE_NAME, {})
    context_config = plugin_config.get(constants.PluginContexts.CONFIG, {})
    project_type_settings = context_config.get(project_type, {})
    return project_type_settings.get(view_name)



"""
Tests for utilities in monitoring.
"""

from unittest.mock import patch

from edx_django_utils.monitoring import background_task


@patch('edx_django_utils.monitoring.internal.utils.newrelic')
def test_background_task_wrapper(wrapped_nr):
    # We are verifying that this returns the correct decorated function
    # in the two cases we care about.
    returned_func = background_task()

    assert returned_func == wrapped_nr.agent.background_task()


@patch('edx_django_utils.monitoring.internal.utils.newrelic', None)
def test_background_task_wrapper_no_new_relic():
    # Test that the decorator behaves as a no-op when newrelic is not set.
    returned_func = background_task()
    wrapped_value = returned_func('a')

    assert wrapped_value == 'a'



"""
Middleware for code_owner custom attribute
"""
import logging

from django.urls import resolve
from django.urls.exceptions import Resolver404

from ..transactions import get_current_transaction
from ..utils import set_custom_attribute
from .utils import (
    _get_catch_all_code_owner,
    get_code_owner_from_module,
    is_code_owner_mappings_configured,
    set_code_owner_custom_attributes
)

log = logging.getLogger(__name__)


class CodeOwnerMonitoringMiddleware:
    """
    Django middleware object to set custom attributes for the owner of each view.

    For instructions on usage, see:
    https://github.com/openedx/edx-django-utils/blob/master/edx_django_utils/monitoring/docs/how_tos/add_code_owner_custom_attribute_to_an_ida.rst

    Custom attributes set:
    - code_owner: The owning team mapped to the current view.
    - code_owner_module: The module found from the request or current transaction.
    - code_owner_path_error: The error mapping by path, if code_owner isn't found in other ways.
    - code_owner_transaction_error: The error mapping by transaction, if code_owner isn't found in other ways.
    - code_owner_transaction_name: The current transaction name used to try to map to code_owner.
        This can be used to find missing mappings.

    """
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        self._set_code_owner_attribute(request)
        return response

    def process_exception(self, request, exception):    # pylint: disable=W0613
        self._set_code_owner_attribute(request)

    def _set_code_owner_attribute(self, request):
        """
        Sets the code_owner custom attribute for the request.
        """
        code_owner = None
        module = self._get_module_from_request(request)
        if module:
            code_owner = get_code_owner_from_module(module)
        if not code_owner:
            code_owner = _get_catch_all_code_owner()

        if code_owner:
            set_code_owner_custom_attributes(code_owner)

    def _get_module_from_request(self, request):
        """
        Get the module from the request path or the current transaction.

        Side-effects:
            Sets code_owner_module custom attribute, used to determine code_owner.
            If module was not found, may set code_owner_path_error and/or
                code_owner_transaction_error custom attributes if applicable.

        Returns:
            str: module name or None if not found

        """
        if not is_code_owner_mappings_configured():
            return None

        module, path_error = self._get_module_from_request_path(request)
        if module:
            set_custom_attribute('code_owner_module', module)
            return module

        module, transaction_error = self._get_module_from_current_transaction()
        if module:
            set_custom_attribute('code_owner_module', module)
            return module

        # monitor errors if module was not found
        if path_error:
            set_custom_attribute('code_owner_path_error', path_error)
        if transaction_error:
            set_custom_attribute('code_owner_transaction_error', transaction_error)
        return None

    def _get_module_from_request_path(self, request):
        """
        Uses the request path to get the view_func module.

        Returns:
            (str, str): (module, error_message), where at least one of these should be None

        """
        try:
            view_func, _, _ = resolve(request.path)
            module = view_func.__module__
            return module, None
        # TODO: Replace ImportError with ModuleNotFoundError when Python 3.5 support is dropped.
        except (ImportError, Resolver404) as e:
            return None, str(e)
        except Exception as e:  # pylint: disable=broad-except; #pragma: no cover
            # will remove broad exceptions after ensuring all proper cases are covered
            set_custom_attribute('deprecated_broad_except__get_module_from_request_path', e.__class__)
            return None, str(e)

    def _get_module_from_current_transaction(self):
        """
        Uses the current transaction to get the module.

        Side-effects:
            Sets code_owner_transaction_name custom attribute, used to determine code_owner

        Returns:
            (str, str): (module, error_message), where at least one of these should be None

        """
        try:
            # Example: openedx.core.djangoapps.contentserver.middleware:StaticContentServer
            transaction_name = get_current_transaction().name
            if not transaction_name:
                return None, 'No current transaction name found.'
            module = transaction_name.split(':')[0]
            set_custom_attribute('code_owner_transaction_name', transaction_name)
            return module, None
        except Exception as e:  # pylint: disable=broad-except
            # will remove broad exceptions after ensuring all proper cases are covered
            set_custom_attribute('deprecated_broad_except___get_module_from_current_transaction', e.__class__)
            return None, str(e)



"""
Tests for the RequestCacheMiddleware.
"""
from unittest.mock import MagicMock

from django.test import RequestFactory, TestCase

from edx_django_utils.cache import middleware
from edx_django_utils.cache.utils import FORCE_CACHE_MISS_PARAM, SHOULD_FORCE_CACHE_MISS_KEY, RequestCache

TEST_KEY = "clobert"
EXPECTED_VALUE = "bertclob"
TEST_NAMESPACE = "test_namespace"


class TestRequestCacheMiddleware(TestCase):  # pylint: disable=missing-class-docstring

    def setUp(self):
        super().setUp()
        self.middleware = middleware.RequestCacheMiddleware()
        self.request = RequestFactory().get('/')

        self.request_cache = RequestCache()
        self.other_request_cache = RequestCache(TEST_NAMESPACE)
        self._dirty_request_cache()

    def test_process_request(self):
        self.middleware.process_request(self.request)

        self._check_request_caches_cleared()

    def test_process_response(self):
        response = self.middleware.process_response(self.request, EXPECTED_VALUE)

        self.assertEqual(response, EXPECTED_VALUE)
        self._check_request_caches_cleared()

    def _check_request_caches_cleared(self):
        """ Checks that all request caches were cleared. """
        self.assertFalse(self.request_cache.get_cached_response(TEST_KEY).is_found)
        self.assertFalse(self.other_request_cache.get_cached_response(TEST_KEY).is_found)

    def _dirty_request_cache(self):
        """ Dirties the request caches to ensure the middleware is clearing it. """
        self.request_cache.set(TEST_KEY, EXPECTED_VALUE)
        self.other_request_cache.set(TEST_KEY, EXPECTED_VALUE)


class TestTieredCacheMiddleware(TestCase):  # pylint: disable=missing-class-docstring

    def setUp(self):
        super().setUp()
        self.middleware = middleware.TieredCacheMiddleware()
        self.request = RequestFactory().get('/')
        self.request.user = self._mock_user(is_staff=True)

        self.request_cache = RequestCache()
        self.request_cache.clear_all_namespaces()

    def test_process_request(self):
        self.middleware.process_request(self.request)

        self.assertFalse(self.request_cache.get_cached_response(SHOULD_FORCE_CACHE_MISS_KEY).value)

    def test_process_request_force_cache_miss(self):
        request = RequestFactory().get(f'/?{FORCE_CACHE_MISS_PARAM}=tRuE')
        request.user = self._mock_user(is_staff=True)

        self.middleware.process_request(request)

        self.assertTrue(self.request_cache.get_cached_response(SHOULD_FORCE_CACHE_MISS_KEY).value)

    def test_process_request_force_cache_miss_non_staff(self):
        request = RequestFactory().get(f'/?{FORCE_CACHE_MISS_PARAM}=tRuE')
        request.user = self._mock_user(is_staff=False)

        self.middleware.process_request(request)

        self.assertFalse(self.request_cache.get_cached_response(SHOULD_FORCE_CACHE_MISS_KEY).value)

    def _mock_user(self, is_staff=True):
        mock_user = MagicMock()
        mock_user.is_active = True
        mock_user.is_staff = is_staff
        return mock_user



"""
Tests for utilities.
"""

from django.test import override_settings

from edx_django_utils.plugins import pluggable_override


@pluggable_override('OVERRIDE_TRANSFORM')
def transform(x):
    return x + 10


def decrement(prev_fn, x):
    if x >= 10:
        return x - 1
    else:
        return prev_fn(x) - 1


def double(prev_fn, x):
    if x >= 11:
        return x * 2
    else:
        return prev_fn(x) * 2


def test_no_override():
    """Test that the original function is called when an override is not specified."""
    assert transform(10) == 20


@override_settings(OVERRIDE_TRANSFORM="{}.decrement".format(__name__))
def test_override():
    """Test that the overriding function is called."""
    assert transform(10) == 9


@override_settings(OVERRIDE_TRANSFORM="{}.decrement".format(__name__))
def test_call_original_function():
    """Test that the overriding function calls the base one."""
    assert transform(9) == 18


@override_settings(OVERRIDE_TRANSFORM="{0}.decrement,{0}.double".format(__name__).split(','))
def test_multiple_overrides_call_last_function():
    """Test that the newest (last) overriding function is called when multiple overrides are specified."""
    assert transform(11) == 22


@override_settings(OVERRIDE_TRANSFORM="{0}.decrement,{0}.double".format(__name__).split(','))
def test_multiple_overrides_fallback_to_previous_function():
    """Test that the last overriding function can call the previous one from the chain."""
    assert transform(10) == 18


@override_settings(OVERRIDE_TRANSFORM="{0}.decrement,{0}.double".format(__name__).split(','))
def test_multiple_overrides_fallback_to_base_function():
    """Test that the overriding functions can eventually call the base one."""
    assert transform(9) == 36



"""
Caching utility middleware.
"""
from django.utils.deprecation import MiddlewareMixin

from . import RequestCache, TieredCache


class RequestCacheMiddleware(MiddlewareMixin):
    """
    Middleware to clear the request cache as appropriate for new requests.
    """
    def process_request(self, request):
        """
        Clears the request cache before processing the request.
        """
        RequestCache.clear_all_namespaces()

    def process_response(self, request, response):
        """
        Clear the request cache after processing a response.
        """
        # This is to just to do some basic cleanup, and isn't
        # necessary for correctness; the next request will have the
        # cache cleared during process_request anyhow.
        #
        # Note that clearing in process_exception would actually cause
        # problems for other middleware that want to read the cache in
        # their process_response.
        RequestCache.clear_all_namespaces()
        return response


class TieredCacheMiddleware(MiddlewareMixin):
    """
    Middleware to store whether or not to force django cache misses.
    """
    def process_request(self, request):
        """
        Stores whether or not FORCE_CACHE_MISS_PARAM was supplied in the
        request.
        """
        TieredCache._get_and_set_force_cache_miss(request)  # pylint: disable=protected-access



"""
This is a collection of transaction-related monitoring utilities.

Usage:

    from edx_django_utils.monitoring import ignore_transaction
    ...
    # called from inside a view you want ignored
    ignore_transaction()

Please remember to expose any new methods in the `__init__.py` file.
"""

from contextlib import contextmanager

try:
    import newrelic.agent
except ImportError:
    newrelic = None  # pylint: disable=invalid-name


def set_monitoring_transaction_name(name, group=None, priority=None):
    """
    Sets the transaction name for monitoring.

    This is not cached, and only support reporting to New Relic.

    """
    if newrelic:  # pragma: no cover
        newrelic.agent.set_transaction_name(name, group, priority)


def ignore_transaction():
    """
    Ignore the transaction in monitoring

    This allows us to ignore code paths that are unhelpful to include, such as
    `/health/` checks.
    """
    if newrelic:  # pragma: no cover
        newrelic.agent.ignore_transaction()


@contextmanager
def function_trace(function_name):
    """
    Wraps a chunk of code that we want to appear as a separate, explicit,
    segment in our monitoring tools.
    """
    # Not covering this because if we mock it, we're not really testing anything
    # anyway. If something did break, it should show up in tests for apps that
    # use this code with newrelic enabled, on whatever version of newrelic they
    # run.
    if newrelic:  # pragma: no cover
        if newrelic.version_info[0] >= 5:
            with newrelic.agent.FunctionTrace(function_name):
                yield
        else:
            nr_transaction = newrelic.agent.current_transaction()
            with newrelic.agent.FunctionTrace(nr_transaction, function_name):
                yield
    else:
        yield


class MonitoringTransaction():
    """
    Represents a monitoring transaction (likely the current transaction).
    """
    def __init__(self, transaction):
        self.transaction = transaction

    @property
    def name(self):
        """
        The name of the transaction.

        For NewRelic, the name may look like:
            openedx.core.djangoapps.contentserver.middleware:StaticContentServer

        """
        if self.transaction and hasattr(self.transaction, 'name'):
            return self.transaction.name
        return None


def get_current_transaction():
    """
    Returns the current transaction.
    """
    current_transaction = None
    if newrelic:
        current_transaction = newrelic.agent.current_transaction()

    return MonitoringTransaction(current_transaction)



"""
Allows plugins to work with django signals

Please remember to expose any new public methods in the `__init__.py` file.
"""
from logging import getLogger

from . import constants, registry, utils

log = getLogger(__name__)


def connect_plugin_receivers(project_type):
    """
    connects receivers to right signal
    """
    for signals_module, signals_config in _iter_plugins(project_type):
        for signal, receiver_func, receiver_config in _iter_receivers(
            signals_module, signals_config
        ):
            signal.connect(
                receiver_func,
                sender=_get_sender(receiver_config),
                dispatch_uid=_get_dispatch_uuid(receiver_config, receiver_func),
            )


def _iter_receivers(signals_module, signals_config):
    """
    Generator for ___ TODO
    """
    for receiver_config in signals_config.get(constants.PluginSignals.RECEIVERS, []):
        receiver_func = utils.import_attr_in_module(
            signals_module, receiver_config[constants.PluginSignals.RECEIVER_FUNC_NAME],
        )
        signal = utils.import_attr(receiver_config[constants.PluginSignals.SIGNAL_PATH])
        yield signal, receiver_func, receiver_config


def _iter_plugins(project_type):
    """
    Generator for ___ TODO
    """
    for app_config in registry.get_plugin_app_configs(project_type):
        signals_config = _get_config(app_config, project_type)
        if signals_config is None:
            log.debug(
                "Plugin Apps [Signals]: Did NOT find %s for %s",
                app_config.name,
                project_type,
            )
            continue

        signals_module_path = utils.get_module_path(
            app_config, signals_config, constants.PluginSignals
        )
        signals_module = utils.import_module(signals_module_path)

        log.debug(
            "Plugin Apps [Signals]: Found %s with %d receiver(s) for %s",
            app_config.name,
            len(signals_config.get(constants.PluginSignals.RECEIVERS, [])),
            project_type,
        )
        yield signals_module, signals_config


def _get_config(app_config, project_type):
    plugin_config = getattr(app_config, constants.PLUGIN_APP_CLASS_ATTRIBUTE_NAME, {})
    signals_config = plugin_config.get(constants.PluginSignals.CONFIG, {})
    return signals_config.get(project_type)


def _get_sender(receiver_config):
    sender_path = receiver_config.get(constants.PluginSignals.SENDER_PATH)
    if sender_path:
        return utils.import_attr(sender_path)
    return None


def _get_dispatch_uuid(receiver_config, receiver_func):
    dispatch_uid = receiver_config.get(constants.PluginSignals.DISPATCH_UID)
    if dispatch_uid is None:
        dispatch_uid = f"{receiver_func.__module__}.{receiver_func.__name__}"
    return dispatch_uid



"""
edx_django_utils Django application initialization.
"""

from django.apps import AppConfig


class EdxDjangoUtilsConfig(AppConfig):
    """
    Configuration for the edx_django_utils Django application.
    """

    name = 'edx_django_utils'



"""
Utilities for working effectively with databases in django.

read_replica: Tools for making queries from the read-replica.
queryset_utils: Utils to use with Django QuerySets.
"""

from .queryset_utils import chunked_queryset
from .read_replica import (
    ReadReplicaRouter,
    read_queries_only,
    read_replica_or_default,
    use_read_replica_if_available,
    write_queries
)



"""
Tests for logging.
"""

from unittest.mock import MagicMock, patch

from django.test import TestCase

from edx_django_utils.logging import RemoteIpFilter, UserIdFilter


class MockRecord:
    """
    Mocks a logging construct to receive data to be interpolated.
    """
    def __init__(self):
        self.userid = None
        self.remoteip = None


class TestLoggingFilters(TestCase):
    """
    Test the logging filters for users and IP addresses
    """

    @patch('edx_django_utils.logging.internal.filters.get_current_user')
    def test_userid_filter(self, mock_get_user):
        mock_user = MagicMock()
        mock_user.pk = '1234'
        mock_get_user.return_value = mock_user

        user_filter = UserIdFilter()
        test_record = MockRecord()
        user_filter.filter(test_record)

        self.assertEqual(test_record.userid, '1234')

    def test_userid_filter_no_user(self):
        user_filter = UserIdFilter()
        test_record = MockRecord()
        user_filter.filter(test_record)

        self.assertEqual(test_record.userid, None)

    @patch('edx_django_utils.logging.internal.filters.get_current_request')
    def test_remoteip_filter(self, mock_get_request):
        mock_request = MagicMock()
        mock_request.META = {'REMOTE_ADDR': '192.168.1.1'}
        mock_get_request.return_value = mock_request

        ip_filter = RemoteIpFilter()
        test_record = MockRecord()
        ip_filter.filter(test_record)

        self.assertEqual(test_record.remoteip, '192.168.1.1')

    def test_remoteip_filter_no_request(self):
        ip_filter = RemoteIpFilter()
        test_record = MockRecord()
        ip_filter.filter(test_record)

        self.assertEqual(test_record.remoteip, None)



"""
Utilities for logging sensitive debug information such as authentication tokens.

Usage:

1. Generate keys using ``log-sensitive gen-keys``
2. Follow the instructions it prints out, and pay close attention to the warning
   at the end of the output
3. When logging sensitive information, use like so::

     logger.info(
         "Received invalid auth token %s in Authorization header",
         encrypt_for_log(token, getattr(settings, 'YOUR_DEBUG_PUBLIC_KEY', None))
     )

   This will log a message like::

     Received invalid auth token [encrypted: ZXI...fFo=|IYS...1KA==] in Authorization header

4. If you need to decrypt one of these messages, save the encrypted portion
   to file, retrieve the securely held private key, and run
   ``log-sensitive decrypt --help`` for instructions.
"""

import sys
from base64 import b64decode, b64encode

import click
from nacl.public import Box, PrivateKey, PublicKey

# Background:
#
# The NaCl "Box" construction provides asymmetric encryption, allowing
# the sender to encrypt something for a recipient without having a
# shared secret. This encryption is authenticated, meaning that the
# recipient verifies that the message matches the sender's public key
# (proof of sender). But it's also *repudiable* authentication; the
# design allows both the sender and the receiver to read (or have
# created!) the encrypted message, so the receipient can't prove to
# anyone *else* that the sender was the author.
#
# Why we use ephemeral sender keys:
#
# The Box is normally an ideal construction to use for
# communications. However, we don't want the logger to be able to read
# the messages it writes, especially not messages from a different
# server instance or from days or weeks ago. Only developers (or
# others) in possession of the recipient keypair should be able to
# read it, not anyone who compromises a server at some later
# date. Luckily, we also don't care about authenticating the logged
# messages as truly being from the server! The solution is for each
# server to create a fresh public/private keypair at startup and to
# include a copy of the public key in any encrypted logs it writes.


# Generate an ephemeral private key for the logger to use during this
# logging session.
logger_private_key = PrivateKey.generate()


def encrypt_for_log(message, reader_public_key_b64):
    """
    Encrypt a message so that it can be logged using the given public key,
    but only read by someone possessing the matching private key. The
    public key is provided in base64.

    A separate keypair should be used for each recipient or purpose.

    Returns a string <sender public key> "|" <ciphertext> wrapped in
    some framing text "[encrypted: ...]"; the inner string can be
    decrypted with decrypt_log_message.

    For ease of use, key may be None or empty; a warning message will be
    returned instead of data, encrypted or otherwise.
    """
    # If no key was configured, don't raise an error. This method is
    # expected to be called from debugging code, so it shouldn't blow
    # up on misconfiguration.
    if not reader_public_key_b64:
        return '[encryption failed, no key]'

    reader_public_key = PublicKey(b64decode(reader_public_key_b64))

    encrypted = Box(logger_private_key, reader_public_key).encrypt(message.encode())

    pubkey = logger_private_key.public_key
    combined = b64encode(bytes(pubkey)).decode() + '|' + b64encode(encrypted).decode()
    # The goal of this framing text is to make it always clear in log
    # messages that the information is encrypted
    return f"[encrypted: {combined}]"


def decrypt_log_message(encrypted_message, reader_private_key_b64):
    """
    Decrypt a message using the private key that has been stored somewhere
    secure and *not* on the server.
    """
    reader_private_key = PrivateKey(b64decode(reader_private_key_b64))
    sender_public_key_data, encrypted_raw = \
        [b64decode(part) for part in encrypted_message.split('|', 1)]
    return Box(reader_private_key, PublicKey(sender_public_key_data)).decrypt(encrypted_raw).decode()


def generate_reader_keys():
    """
    Utility method for generating a public/private keypair for use with these
    logging functions.
    """
    reader_private_key = PrivateKey.generate()
    return {
        'public': b64encode(bytes(reader_private_key.public_key)).decode(),
        'private': b64encode(bytes(reader_private_key)).decode(),
    }


@click.group()
def cli():
    pass


@click.command('gen-keys', help="Generate keypair")
def cli_gen_keys():  # pragma: no cover
    """
    Generate and print a keypair for handling sensitive log messages.
    """
    reader_keys = generate_reader_keys()
    public_64 = reader_keys['public']
    private_64 = reader_keys['private']
    print(
        "The following is your PUBLIC key, which should be included in the server's "
        "configuration. Create a separate setting (and keypair) for each "
        "distinct project or team. This value does not need special protection:"
        "\n\n"
        f"  settings.<YOUR_DEBUG_PUBLIC_KEY> = \"{public_64}\""
        "\n\n"
        "The following is your PRIVATE key, which must never be present on the server "
        "and should instead be kept encrypted in a separate, safe place "
        "such as a password manager:"
        "\n\n"
        f"  \"{private_64}\" (private)"
        "\n\n"
        "WARNING: Before logging anything sensitive, get a legal/compliance review to "
        "ensure this is acceptable in your organization. Encryption is not "
        "generally a replacement for retention policies or other privacy "
        "safeguards; using this utility does not automatically make sensitive "
        "information safe to handle."
    )


@click.command('decrypt', help="""Decrypt a logged message.

If possible, use bash process indirection to keep the private key from
touching disk or shell history unencrypted. The safest way is to keep
the private key in an encrypted file:

    --private-key-file <(gpg2 --decrypt auth-logging-key.enc)

Alternatively, you could copy it from a password manager to your
clipboard and use a CLI clipboard tool to retrieve it:

\b
    --private-key-file <(xsel -bo) # Linux
    --private-key-file <(pbpaste) # Mac

Another option is to somehow get the private key into an environment
variable and echo it out:

    --private-key-file <(echo "$PRIVATE_KEY")

The same techniques can also be used for the encrypted message data,
which is less sensitive but should also be handled with care.
""")
@click.option(
    '--private-key-file', type=click.File('r'), required=True,
    help="Path to file containing reader's private key in Base64",
)
@click.option(
    '--message-file', type=click.File('r'), required=True,
    help="Path to file containing encrypted message, or - for stdin",
)
def cli_decrypt(private_key_file, message_file):  # pragma: no cover
    """
    Decrypt a message and print it to stdout.
    """
    message = message_file.read()
    if 'encrypted:' in message:
        print("ERROR: Only include the Base64, not the 'encrypted:' wrapper.", file=sys.stderr)
        sys.exit(1)
    print(decrypt_log_message(message, private_key_file.read()))


@click.command('encrypt', help="Encrypt a one-off message (for testing)")
@click.option('--public-key', help="Reader's public key, in Base64")
@click.option(
    '--message-file', type=click.File('r'), required=True,
    help="Path to file containing message to encrypt, or - for stdin",
)
def cli_encrypt(public_key, message_file):  # pragma: no cover
    """
    Encrypt a message to the provided public key and print it to stdout.

    This is just intended for use when testing or experimenting with the decrypt command.
    """
    print(encrypt_for_log(message_file.read(), public_key))


cli.add_command(cli_gen_keys)
cli.add_command(cli_decrypt)
cli.add_command(cli_encrypt)

if __name__ == '__main__':
    cli()



"""
Tests for CachedCustomMonitoringMiddleware and associated utilities.

Note: See test_middleware.py for the rest of the middleware tests.
"""
from unittest.mock import call, patch

import ddt
from django.test import TestCase

from edx_django_utils.cache import RequestCache
from edx_django_utils.monitoring import (
    CachedCustomMonitoringMiddleware,
    accumulate,
    get_current_transaction,
    increment,
    record_exception
)

from ..middleware import CachedCustomMonitoringMiddleware as DeprecatedCachedCustomMonitoringMiddleware
from ..middleware import MonitoringCustomMetricsMiddleware as DeprecatedMonitoringCustomMetricsMiddleware
from ..utils import accumulate as deprecated_accumulate
from ..utils import increment as deprecated_increment
from ..utils import set_custom_attribute as deprecated_set_custom_attribute
from ..utils import set_custom_attributes_for_course_key as deprecated_set_custom_attributes_for_course_key


@ddt.ddt
class TestCustomMonitoringMiddleware(TestCase):
    """
    Test the monitoring_utils middleware and helpers
    """
    def setUp(self):
        super().setUp()
        RequestCache.clear_all_namespaces()

    @patch('newrelic.agent')
    @ddt.data(
        (CachedCustomMonitoringMiddleware, False, 'process_response'),
        (CachedCustomMonitoringMiddleware, False, 'process_exception'),
        (DeprecatedCachedCustomMonitoringMiddleware, True, 'process_response'),
        (DeprecatedMonitoringCustomMetricsMiddleware, True, 'process_response'),
    )
    @ddt.unpack
    def test_accumulate_and_increment(
            self, cached_monitoring_middleware_class, is_deprecated, middleware_method_name, mock_newrelic_agent
    ):
        """
        Test normal usage of collecting custom attributes and reporting to New Relic
        """
        accumulate('hello', 10)
        accumulate('world', 10)
        accumulate('world', 10)
        increment('foo')
        increment('foo')

        # based on the attribute data above, we expect the following calls to newrelic:
        nr_agent_calls_expected = [
            call('hello', 10),
            call('world', 20),
            call('foo', 2),
        ]

        # fake a response to trigger attributes reporting
        middleware_method = getattr(cached_monitoring_middleware_class(), middleware_method_name)
        middleware_method(
            'fake request',
            'fake response',
        )

        # Assert call counts to newrelic.agent.add_custom_parameter()
        expected_call_count = len(nr_agent_calls_expected)
        if is_deprecated:
            expected_call_count += 1
        measured_call_count = mock_newrelic_agent.add_custom_parameter.call_count
        self.assertEqual(expected_call_count, measured_call_count)

        # Assert call args to newrelic.agent.add_custom_parameter().  Due to
        # the nature of python dicts, call order is undefined.
        mock_newrelic_agent.add_custom_parameter.assert_has_calls(nr_agent_calls_expected, any_order=True)

    @patch('newrelic.agent')
    @ddt.data(
        (CachedCustomMonitoringMiddleware, False),
        (DeprecatedCachedCustomMonitoringMiddleware, True),
        (DeprecatedMonitoringCustomMetricsMiddleware, True),
    )
    @ddt.unpack
    def test_accumulate_with_illegal_value(
            self, cached_monitoring_middleware_class, is_deprecated, mock_newrelic_agent
    ):
        """
        Test monitoring accumulate with illegal value that can't be added.
        """
        accumulate('hello', None)
        accumulate('hello', 10)

        # based on the metric data above, we expect the following calls to newrelic:
        nr_agent_calls_expected = [
            call('hello', None),
            call('error_adding_accumulated_metric', 'name=hello, new_value=10, cached_value=None'),
        ]

        # fake a response to trigger metrics reporting
        cached_monitoring_middleware_class().process_response(
            'fake request',
            'fake response',
        )

        # Assert call counts to newrelic.agent.add_custom_parameter()
        expected_call_count = len(nr_agent_calls_expected)
        if is_deprecated:
            expected_call_count += 1
        measured_call_count = mock_newrelic_agent.add_custom_parameter.call_count
        self.assertEqual(expected_call_count, measured_call_count)

        # Assert call args to newrelic.agent.add_custom_parameter().
        mock_newrelic_agent.add_custom_parameter.assert_has_calls(nr_agent_calls_expected, any_order=True)

    @patch('newrelic.agent')
    def test_get_current_transaction(self, mock_newrelic_agent):
        mock_newrelic_agent.current_transaction().name = 'fake-transaction'
        current_transaction = get_current_transaction()
        self.assertEqual(current_transaction.name, 'fake-transaction')

    def test_get_current_transaction_without_newrelic(self):
        current_transaction = get_current_transaction()
        self.assertEqual(current_transaction.name, None)

    @patch('edx_django_utils.monitoring.utils.internal_accumulate')
    def test_deprecated_accumulate(self, mock_accumulate):
        deprecated_accumulate('key', 1)
        mock_accumulate.assert_called_with('key', 1)

    @patch('edx_django_utils.monitoring.utils.internal_increment')
    def test_deprecated_increment(self, mock_increment):
        deprecated_increment('key')
        mock_increment.assert_called_with('key')

    @patch('edx_django_utils.monitoring.utils.internal_set_custom_attribute')
    def test_deprecated_set_custom_attribute(self, mock_set_custom_attribute):
        deprecated_set_custom_attribute('key', True)
        mock_set_custom_attribute.assert_called_with('key', True)

    @patch('edx_django_utils.monitoring.utils.internal_set_custom_attributes_for_course_key')
    def test_deprecated_set_custom_attributes_for_course_key(self, mock_set_custom_attributes_for_course_key):
        deprecated_set_custom_attributes_for_course_key('key')
        mock_set_custom_attributes_for_course_key.assert_called_with('key')

    @patch('newrelic.agent.record_exception')
    def test_record_exception(self, mock_record_exception):
        record_exception()
        mock_record_exception.assert_called_once()



"""
Deprecated Middleware for backward-compatibility.

IMPORTANT: No new classes should be added to this file.
TODO: Remove this file once these classes are no longer used.

"""
import warnings

from .internal.middleware import CachedCustomMonitoringMiddleware as InternalCachedCustomMonitoringMiddleware
from .internal.middleware import MonitoringMemoryMiddleware as InternalMonitoringMemoryMiddleware
from .internal.utils import set_custom_attribute


class CachedCustomMonitoringMiddleware(InternalCachedCustomMonitoringMiddleware):
    """
    Deprecated class for handling middleware. Class has been moved to public API.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        msg = "Use 'edx_django_utils.monitoring.CachedCustomMonitoringMiddleware' in place of " \
              "'edx_django_utils.monitoring.middleware.CachedCustomMonitoringMiddleware'."
        warnings.warn(msg, DeprecationWarning)
        set_custom_attribute('deprecated_monitoring_middleware', 'CachedCustomMonitoringMiddleware')


class MonitoringCustomMetricsMiddleware(InternalCachedCustomMonitoringMiddleware):
    """
    Deprecated class for handling middleware. Class has been renamed to CachedCustomMonitoringMiddleware.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        msg = "Use 'edx_django_utils.monitoring.CachedCustomMonitoringMiddleware' in place of " \
              "'edx_django_utils.monitoring.middleware.MonitoringCustomMetricsMiddleware'."
        warnings.warn(msg, DeprecationWarning)
        set_custom_attribute('deprecated_monitoring_middleware', 'MonitoringCustomMetricsMiddleware')


class MonitoringMemoryMiddleware(InternalMonitoringMemoryMiddleware):
    """
    Deprecated class for handling middleware. Class has been moved to public API.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        msg = "Use 'edx_django_utils.monitoring.MonitoringMemoryMiddleware' in place of " \
              "'edx_django_utils.monitoring.middleware.MonitoringMemoryMiddleware'."
        warnings.warn(msg, DeprecationWarning)
        set_custom_attribute('deprecated_monitoring_middleware', 'MonitoringMemoryMiddleware')



"""
Unit tests for user_management management commands.
"""


import sys

import ddt
import pytest
from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.core.management import CommandError, call_command
from django.test import TestCase

TEST_EMAIL = 'test@example.com'
TEST_GROUP = 'test-group'
TEST_USERNAME = 'test-user'
TEST_DATA = (
    {},
    {
        TEST_GROUP: ['add_group', 'change_group', 'change_group'],
    },
    {
        'other-group': ['add_group', 'change_group', 'change_group'],
    },
)


@ddt.ddt
class TestManageGroupCommand(TestCase):
    """
    Tests the `manage_group` command.
    """

    def set_group_permissions(self, group_permissions):
        """
        Sets up a before-state for groups and permissions in tests, which
        can be checked afterward to ensure that a failed atomic
        operation has not had any side effects.
        """
        content_type = ContentType.objects.get_for_model(Group)
        for group_name, permission_codenames in group_permissions.items():
            group = Group.objects.create(name=group_name)
            for codename in permission_codenames:
                group.permissions.add(
                    Permission.objects.get(content_type=content_type, codename=codename)
                )

    def check_group_permissions(self, group_permissions):
        """
        Checks that the current state of the database matches the specified groups and
        permissions.
        """
        self.check_groups(list(group_permissions.keys()))
        for group_name, permission_codenames in group_permissions.items():
            self.check_permissions(group_name, permission_codenames)

    def check_groups(self, group_names):
        """
        DRY helper.
        """
        assert set(group_names) == {g.name for g in Group.objects.all()}

    def check_permissions(self, group_name, permission_codenames):
        """
        DRY helper.
        """
        assert set(permission_codenames) == {p.codename for p in Group.objects.get(name=group_name).permissions.all()}

    @ddt.data(
        *(
            (data, args, exception)
            for data in TEST_DATA
            for args, exception in (
                ((), 'too few arguments' if sys.version_info.major == 2 else 'required: group_name'),  # no group name
                (('x' * 151,), 'invalid group name'),  # invalid group name
                ((TEST_GROUP, 'some-other-group'), 'unrecognized arguments'),  # multiple arguments
                ((TEST_GROUP, '--some-option', 'dummy'), 'unrecognized arguments')  # unexpected option name
            )
        )
    )
    @ddt.unpack
    def test_invalid_input(self, initial_group_permissions, command_args, exception_message):
        """
        Ensures that invalid inputs result in errors with relevant output,
        and that no persistent state is changed.
        """
        self.set_group_permissions(initial_group_permissions)

        with pytest.raises(CommandError) as exc_context:
            call_command('manage_group', *command_args)
        assert exception_message in str(exc_context.value).lower()
        self.check_group_permissions(initial_group_permissions)

    @ddt.data(*TEST_DATA)
    def test_invalid_permission(self, initial_group_permissions):
        """
        Ensures that a permission that cannot be parsed or resolved results in
        and error and that no persistent state is changed.
        """
        self.set_group_permissions(initial_group_permissions)

        # not parseable
        with pytest.raises(CommandError) as exc_context:
            call_command('manage_group', TEST_GROUP, '--permissions', 'fail')
        assert 'invalid permission option' in str(exc_context.value).lower()
        self.check_group_permissions(initial_group_permissions)

        # not parseable
        with pytest.raises(CommandError) as exc_context:
            call_command('manage_group', TEST_GROUP, '--permissions', 'f:a:i:l')
        assert 'invalid permission option' in str(exc_context.value).lower()
        self.check_group_permissions(initial_group_permissions)

        # invalid app label
        with pytest.raises(CommandError) as exc_context:
            call_command('manage_group', TEST_GROUP, '--permissions', 'nonexistent-label:dummy-model:dummy-perm')
        assert 'no installed app' in str(exc_context.value).lower()
        assert 'nonexistent-label' in str(exc_context.value).lower()
        self.check_group_permissions(initial_group_permissions)

        # invalid model name
        with pytest.raises(CommandError) as exc_context:
            call_command('manage_group', TEST_GROUP, '--permissions', 'auth:nonexistent-model:dummy-perm')
        assert 'nonexistent-model' in str(exc_context.value).lower()
        self.check_group_permissions(initial_group_permissions)

        # invalid model name
        with pytest.raises(CommandError) as exc_context:
            call_command('manage_group', TEST_GROUP, '--permissions', 'auth:Group:nonexistent-perm')
        assert 'invalid permission codename' in str(exc_context.value).lower()
        assert 'nonexistent-perm' in str(exc_context.value).lower()
        self.check_group_permissions(initial_group_permissions)

    def test_group(self):
        """
        Ensures that groups are created if they don't exist and reused if they do.
        """
        self.check_groups([])
        call_command('manage_group', TEST_GROUP)
        self.check_groups([TEST_GROUP])

        # check idempotency
        call_command('manage_group', TEST_GROUP)
        self.check_groups([TEST_GROUP])

    def test_group_remove(self):
        """
        Ensures that groups are removed if they exist and we exit cleanly otherwise.
        """
        self.set_group_permissions({TEST_GROUP: ['add_group']})
        self.check_groups([TEST_GROUP])
        call_command('manage_group', TEST_GROUP, '--remove')
        self.check_groups([])

        # check idempotency
        call_command('manage_group', TEST_GROUP, '--remove')
        self.check_groups([])

    def test_permissions(self):
        """
        Ensures that permissions are set on the group as specified.
        """
        self.check_groups([])
        call_command('manage_group', TEST_GROUP, '--permissions', 'auth:Group:add_group')
        self.check_groups([TEST_GROUP])
        self.check_permissions(TEST_GROUP, ['add_group'])

        # check idempotency
        call_command('manage_group', TEST_GROUP, '--permissions', 'auth:Group:add_group')
        self.check_groups([TEST_GROUP])
        self.check_permissions(TEST_GROUP, ['add_group'])

        # check adding a permission
        call_command('manage_group', TEST_GROUP, '--permissions', 'auth:Group:add_group', 'auth:Group:change_group')
        self.check_groups([TEST_GROUP])
        self.check_permissions(TEST_GROUP, ['add_group', 'change_group'])

        # check removing a permission
        call_command('manage_group', TEST_GROUP, '--permissions', 'auth:Group:change_group')
        self.check_groups([TEST_GROUP])
        self.check_permissions(TEST_GROUP, ['change_group'])

        # check removing all permissions
        call_command('manage_group', TEST_GROUP)
        self.check_groups([TEST_GROUP])
        self.check_permissions(TEST_GROUP, [])



"""
Constants/Enums used during interfacing with plugin system
"""

# Name of the class attribute to put in the AppConfig class of the Plugin App.
PLUGIN_APP_CLASS_ATTRIBUTE_NAME = "plugin_app"


# Name of the function that belongs in the plugin Django app's settings file.
# The function should be defined as:
#     def plugin_settings(settings):
#         # enter code that should be injected into the given settings module.
PLUGIN_APP_SETTINGS_FUNC_NAME = "plugin_settings"


class PluginSettings():
    """
    The PluginSettings enum defines dictionary field names (and defaults)
    that can be specified by a Plugin App in order to configure the settings
    that are injected into the project.
    """

    CONFIG = "settings_config"
    RELATIVE_PATH = "relative_path"
    DEFAULT_RELATIVE_PATH = "settings"


class PluginURLs():
    """
    The PluginURLs enum defines dictionary field names (and defaults) that can
    be specified by a Plugin App in order to configure the URLs that are
    injected into the project.
    """

    CONFIG = "url_config"
    APP_NAME = "app_name"
    NAMESPACE = "namespace"
    REGEX = "regex"
    RELATIVE_PATH = "relative_path"
    DEFAULT_RELATIVE_PATH = "urls"


class PluginSignals():
    """
    The PluginSignals enum defines dictionary field names (and defaults)
    that can be specified by a Plugin App in order to configure the signals
    that it receives.
    """

    CONFIG = "signals_config"

    RECEIVERS = "receivers"
    DISPATCH_UID = "dispatch_uid"
    RECEIVER_FUNC_NAME = "receiver_func_name"
    SENDER_PATH = "sender_path"
    SIGNAL_PATH = "signal_path"

    RELATIVE_PATH = "relative_path"
    DEFAULT_RELATIVE_PATH = "signals"


class PluginContexts():
    """
    The PluginContexts enum defines dictionary field names (and defaults)
    that can be specified by a Plugin App in order to configure the
    additional views it would like to add context into.
    """

    CONFIG = "view_context_config"



"""
Adds support for first class plugins that can be added to an IDA.

Please remember to expose any new public methods in the `__init__.py` file.
"""

import functools
from collections import OrderedDict

from stevedore.extension import ExtensionManager


class PluginError(Exception):
    """
    Base Exception for when an error was found regarding plugins.
    """


class PluginManager:
    """
    Base class that manages plugins for an IDA.
    """

    @classmethod
    @functools.lru_cache(maxsize=None)
    def get_available_plugins(cls, namespace=None):
        """
        Returns a dict of all the plugins that have been made available.
        """
        # Note: we're creating the extension manager lazily to ensure that the Python path
        # has been correctly set up. Trying to create this statically will fail, unfortunately.
        plugins = OrderedDict()
        # pylint: disable=no-member
        extension_manager = ExtensionManager(
            namespace=namespace or cls.NAMESPACE
        )
        for plugin_name in extension_manager.names():
            plugins[plugin_name] = extension_manager[plugin_name].plugin
        return plugins

    @classmethod
    def get_plugin(cls, name, namespace=None):
        """
        Returns the plugin with the given name.
        """
        plugins = cls.get_available_plugins(namespace)
        if name not in plugins:
            raise PluginError(
                "No such plugin {name} for entry point {namespace}".format(
                    name=name,
                    namespace=namespace or cls.NAMESPACE,  # pylint: disable=no-member
                )
            )
        return plugins[name]



"""
User and group utilities useful to multiple django services will live here.
"""

import random
import string


def generate_password(length=12, chars=string.ascii_letters + string.digits):
    """Generate a valid random password"""
    if length < 8:
        raise ValueError("password must be at least 8 characters")

    choice = random.SystemRandom().choice

    password = ''
    password += choice(string.digits)
    password += choice(string.ascii_letters)
    password += ''.join([choice(chars) for _i in range(length - 2)])
    return password



"""
Tests for the request cache.
"""

from threading import Thread
from unittest import TestCase, mock

import ddt

from edx_django_utils.cache.utils import (
    DEFAULT_REQUEST_CACHE_NAMESPACE,
    SHOULD_FORCE_CACHE_MISS_KEY,
    CachedResponse,
    CachedResponseError,
    RequestCache,
    TieredCache,
    get_cache_key
)

TEST_KEY = "clöbert"
TEST_KEY_2 = "clöbert2"
TEST_KEY_UNICODE = "clöbert"
EXPECTED_VALUE = "bertclöb"
EXPECTED_VALUE_2 = "bertclöb2"
TEST_NAMESPACE = "test_namespåce"
TEST_DJANGO_TIMEOUT_CACHE = 1


class TestRequestCache(TestCase):  # pylint: disable=missing-class-docstring
    def setUp(self):
        super().setUp()
        RequestCache.clear_all_namespaces()
        self.request_cache = RequestCache()
        self.other_request_cache = RequestCache(TEST_NAMESPACE)

    def test_get_cached_response_hit(self):
        self.request_cache.set(TEST_KEY, EXPECTED_VALUE)
        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertTrue(cached_response.is_found)
        self.assertEqual(cached_response.value, EXPECTED_VALUE)
        cached_response = self.other_request_cache.get_cached_response(TEST_KEY)
        self.assertFalse(cached_response.is_found)

        self.other_request_cache.set(TEST_KEY_2, EXPECTED_VALUE)
        cached_response = self.request_cache.get_cached_response(TEST_KEY_2)
        self.assertFalse(cached_response.is_found)
        cached_response = self.other_request_cache.get_cached_response(TEST_KEY_2)
        self.assertTrue(cached_response.is_found)
        self.assertEqual(cached_response.value, EXPECTED_VALUE)

    def test_get_cached_response_hit_with_cached_none(self):
        self.request_cache.set(TEST_KEY, None)
        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertTrue(cached_response.is_found)
        self.assertEqual(cached_response.value, None)

    def test_get_cached_response_miss(self):
        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertFalse(cached_response.is_found)

    def test_get_cached_response_with_default(self):
        self.request_cache.setdefault(TEST_KEY, EXPECTED_VALUE)
        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertTrue(cached_response.is_found)
        self.assertEqual(cached_response.value, EXPECTED_VALUE)

    def test_get_cached_response_with_default_after_set(self):
        self.request_cache.set(TEST_KEY, EXPECTED_VALUE_2)
        self.request_cache.setdefault(TEST_KEY, EXPECTED_VALUE)
        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertTrue(cached_response.is_found)
        self.assertEqual(cached_response.value, EXPECTED_VALUE_2)

    def test_cache_data(self):
        self.assertDictEqual(self.request_cache.data, {})

        key_value_pairs = [
            (TEST_KEY, EXPECTED_VALUE),
            (TEST_KEY_2, EXPECTED_VALUE_2)
        ]
        expected_dict = {}
        for key, value in key_value_pairs:
            self.request_cache.set(key, value)
            expected_dict[key] = value

        self.assertDictEqual(self.request_cache.data, expected_dict)

    def test_clear(self):
        self.request_cache.set(TEST_KEY, EXPECTED_VALUE)
        self.other_request_cache.set(TEST_KEY, EXPECTED_VALUE)
        self.request_cache.clear()
        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertFalse(cached_response.is_found)
        cached_response = self.other_request_cache.get_cached_response(TEST_KEY)
        self.assertTrue(cached_response.is_found)

    def test_clear_all_namespaces(self):
        self.request_cache.set(TEST_KEY, EXPECTED_VALUE)
        self.other_request_cache.set(TEST_KEY, EXPECTED_VALUE)
        RequestCache.clear_all_namespaces()
        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertFalse(cached_response.is_found)

        cached_response = self.other_request_cache.get_cached_response(TEST_KEY)
        self.assertFalse(cached_response.is_found)

    def test_clear_all_namespaces_other_thread(self):
        """
        Clearing all namespaces for a different thread should not clear this
        request cache.
        """
        self.request_cache.set(TEST_KEY, EXPECTED_VALUE)
        other_thread = Thread(target=lambda: RequestCache.clear_all_namespaces())  # pylint: disable=unnecessary-lambda
        other_thread.start()
        other_thread.join()

        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertTrue(cached_response.is_found)

    def test_delete(self):
        self.request_cache.set(TEST_KEY, EXPECTED_VALUE)
        self.request_cache.set(TEST_KEY_2, EXPECTED_VALUE)
        self.other_request_cache.set(TEST_KEY, EXPECTED_VALUE)
        self.request_cache.delete(TEST_KEY)

        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertFalse(cached_response.is_found)
        cached_response = self.request_cache.get_cached_response(TEST_KEY_2)
        self.assertTrue(cached_response.is_found)
        self.assertEqual(cached_response.value, EXPECTED_VALUE)
        cached_response = self.other_request_cache.get_cached_response(TEST_KEY)
        self.assertTrue(cached_response.is_found)

    def test_delete_missing_key(self):
        try:
            self.request_cache.delete(TEST_KEY)
        except KeyError:
            self.fail('Deleting a missing key from the request cache should not cause an error.')

    def test_create_request_cache_with_default_namespace(self):
        with self.assertRaises(AssertionError):
            RequestCache(DEFAULT_REQUEST_CACHE_NAMESPACE)


class TestTieredCache(TestCase):  # pylint: disable=missing-class-docstring
    def setUp(self):
        super().setUp()
        self.request_cache = RequestCache()
        TieredCache.dangerous_clear_all_tiers()

    def test_get_cached_response_all_tier_miss(self):
        cached_response = TieredCache.get_cached_response(TEST_KEY)
        self.assertFalse(cached_response.is_found)

    def test_get_cached_response_request_cache_hit(self):
        self.request_cache.set(TEST_KEY, EXPECTED_VALUE)
        cached_response = TieredCache.get_cached_response(TEST_KEY)
        self.assertTrue(cached_response.is_found)
        self.assertEqual(cached_response.value, EXPECTED_VALUE)

    @mock.patch('django.core.cache.cache.get')
    def test_get_cached_response_django_cache_hit(self, mock_cache_get):
        mock_cache_get.return_value = EXPECTED_VALUE
        cached_response = TieredCache.get_cached_response(TEST_KEY)
        self.assertTrue(cached_response.is_found)
        self.assertEqual(cached_response.value, EXPECTED_VALUE)

        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertTrue(cached_response.is_found, 'Django cache hit should cache value in request cache.')

    @mock.patch('django.core.cache.cache.get')
    def test_get_cached_response_force_cache_miss(self, mock_cache_get):
        self.request_cache.set(SHOULD_FORCE_CACHE_MISS_KEY, True)
        mock_cache_get.return_value = EXPECTED_VALUE
        cached_response = TieredCache.get_cached_response(TEST_KEY)
        self.assertFalse(cached_response.is_found)

        cached_response = self.request_cache.get_cached_response(TEST_KEY)
        self.assertFalse(cached_response.is_found, 'Forced Django cache miss should not cache value in request cache.')

    @mock.patch('django.core.cache.cache.set')
    def test_set_all_tiers(self, mock_cache_set):
        mock_cache_set.return_value = EXPECTED_VALUE
        TieredCache.set_all_tiers(TEST_KEY, EXPECTED_VALUE, TEST_DJANGO_TIMEOUT_CACHE)
        mock_cache_set.assert_called_with(TEST_KEY, EXPECTED_VALUE, TEST_DJANGO_TIMEOUT_CACHE)
        self.assertEqual(self.request_cache.get_cached_response(TEST_KEY).value, EXPECTED_VALUE)

    @mock.patch('django.core.cache.cache.clear')
    def test_dangerous_clear_all_tiers_and_namespaces(self, mock_cache_clear):
        TieredCache.set_all_tiers(TEST_KEY, EXPECTED_VALUE)
        TieredCache.dangerous_clear_all_tiers()
        self.assertFalse(self.request_cache.get_cached_response(TEST_KEY).is_found)
        mock_cache_clear.assert_called_once_with()

    @mock.patch('django.core.cache.cache.delete')
    def test_delete(self, mock_cache_delete):
        TieredCache.set_all_tiers(TEST_KEY, EXPECTED_VALUE)
        TieredCache.set_all_tiers(TEST_KEY_2, EXPECTED_VALUE)
        TieredCache.delete_all_tiers(TEST_KEY)
        self.assertFalse(self.request_cache.get_cached_response(TEST_KEY).is_found)
        self.assertEqual(self.request_cache.get_cached_response(TEST_KEY_2).value, EXPECTED_VALUE)
        mock_cache_delete.assert_called_with(TEST_KEY)


class CacheResponseTests(TestCase):  # pylint: disable=missing-class-docstring
    def test_is_miss(self):
        is_found = False
        cached_response = CachedResponse(is_found, TEST_KEY, EXPECTED_VALUE)
        self.assertFalse(cached_response.is_found)
        self.assertEqual(cached_response.key, TEST_KEY)
        with self.assertRaises(AttributeError):
            cached_response.value  # pylint: disable=pointless-statement
        self.assertEqual(cached_response.get_value_or_default(EXPECTED_VALUE_2), EXPECTED_VALUE_2)
        self.assertIn(
            'CachedResponse(is_found={}, key={}'.format(False, TEST_KEY),
            cached_response.__repr__()  # pylint: disable=unnecessary-dunder-call
        )

    def test_is_hit(self):
        is_found = True
        cached_response = CachedResponse(is_found, TEST_KEY, EXPECTED_VALUE)
        self.assertTrue(cached_response.is_found)
        self.assertEqual(cached_response.key, TEST_KEY)
        self.assertEqual(cached_response.value, EXPECTED_VALUE)
        self.assertEqual(cached_response.get_value_or_default(EXPECTED_VALUE_2), EXPECTED_VALUE)
        self.assertIn(
            'CachedResponse(is_found={}, key={}'.format(True, TEST_KEY),
            cached_response.__repr__()  # pylint: disable=unnecessary-dunder-call
        )

    def test_cached_response_equals(self):
        self.assertEqual(
            CachedResponse(True, TEST_KEY, EXPECTED_VALUE),
            CachedResponse(True, TEST_KEY, EXPECTED_VALUE),
        )
        self.assertEqual(
            CachedResponse(False, TEST_KEY, EXPECTED_VALUE),
            CachedResponse(False, TEST_KEY, EXPECTED_VALUE),
        )

        self.assertNotEqual(
            CachedResponse(True, TEST_KEY, EXPECTED_VALUE),
            CachedResponse(False, TEST_KEY, EXPECTED_VALUE),
        )
        self.assertNotEqual(
            CachedResponse(True, TEST_KEY, EXPECTED_VALUE),
            CachedResponse(True, TEST_KEY_2, EXPECTED_VALUE),
        )
        self.assertNotEqual(
            CachedResponse(True, TEST_KEY, EXPECTED_VALUE),
            CachedResponse(True, TEST_KEY, EXPECTED_VALUE_2),
        )

    def test_cached_response_not_equals(self):
        self.assertNotEqual(
            CachedResponse(True, TEST_KEY, EXPECTED_VALUE), CachedResponse(True, TEST_KEY, EXPECTED_VALUE_2)
        )

    def test_cached_response_misuse(self):
        cached_response = CachedResponse(True, TEST_KEY, EXPECTED_VALUE)

        with self.assertRaises(CachedResponseError):
            bool(cached_response)

        with self.assertRaises(CachedResponseError):
            other_object = object()
            cached_response == other_object  # pylint: disable=pointless-statement


@ddt.ddt
class TestCacheUtils(TestCase):

    """
        Class to test cache utils.
    """
    @ddt.data(
        (get_cache_key(site_domain="example.com", resource="catalogs"), True),   # test same key
        (get_cache_key(resource="catalogs", site_domain="example.com"), True),   # test sorted arguments
        (get_cache_key(site_domain=b"example.com", resource="catalogs"), True),  # test byte str value
        (get_cache_key(site_domain="example.com", resource="users"), False),     # test one different argument
        (get_cache_key(site_domain="other.com", resource="catalogs"), False),    # test another different argument
    )
    @ddt.unpack
    def test_get_cache_key(self, test_key, result):
        """
            tests get_cache_key utility.
        """
        key = get_cache_key(site_domain="example.com", resource="catalogs")
        self.assertEqual(key == test_key, result)



"""
Tests for admin mixins.
"""
import pytest
from django.contrib.admin import ModelAdmin
from django.contrib.admin.sites import AdminSite

from edx_django_utils.admin.mixins import ReadOnlyAdminMixin
from edx_django_utils.admin.tests.models import GenericModel


class ReadOnlyAdmin(ReadOnlyAdminMixin, ModelAdmin):
    """
    Test admin interface which adds the mixin to be tested.
    """
    pass


@pytest.fixture
def site():
    return AdminSite()


@pytest.fixture
def a_request():
    class MockRequest:
        GET = []
    return MockRequest()


@pytest.fixture
def read_only_admin(site):
    return ReadOnlyAdmin(GenericModel, site)


class TestReadOnlyAdminMixin:
    def test_create(self, read_only_admin, a_request):
        assert not read_only_admin.has_add_permission(a_request)

    def test_delete(self, read_only_admin, a_request):
        assert not read_only_admin.has_delete_permission(a_request)

    def test_delete_action(self, read_only_admin, a_request):
        actions = read_only_admin.get_actions(a_request)
        assert 'delete_selected' not in actions

    def test_list_display_links(self, read_only_admin):
        assert read_only_admin.list_display_links is None

    def test_read_only_fields(self, read_only_admin):
        # The GenericModel has three fields plus an Django-added id field - ensure *all* are marked as read-only.
        assert len(read_only_admin.readonly_fields) == 4



"""
Middleware to add Content-Security-Policy and related headers.

The ``content_security_policy_middleware`` middleware function can add
``Content-Security-Policy``, ``Content-Security-Policy-Report-Only``, and
``Reporting-Endpoints`` HTTP response headers. This functionality is
configured by Django settings.

This middleware only supports static values for the headers. That is, it does not
support the ``strict-dynamic`` directive, which requires coordination of nonces or
hashes between the response header and response body.
"""
import re

from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed


def _load_headers() -> dict:
    """
    Return a dict of headers to append to every response, based on settings.
    """
    # .. setting_name: CSP_STATIC_ENFORCE
    # .. setting_default: None
    # .. setting_description: Content-Security-Policy header to attach to all responses.
    #   This should include everything but the ``report-to`` or ``report-uri`` clauses; those
    #   will be appended automatically according to the ``CSP_STATIC_REPORTING_NAME`` and
    #   ``CSP_STATIC_REPORTING_URI`` settings. Newlines are permitted and will be replaced with spaces.
    #   A trailing `;` is also permitted.
    # .. setting_warning: Setting the CSP header to too strict a value can cause your pages to
    #   break. It is strongly recommended that deployers start by using ``CSP_STATIC_REPORT_ONLY`` (along
    #   with the reporting settings) and only move or copy the policies into ``CSP_STATIC_ENFORCE`` after
    #   confirming that the received CSP reports only represent false positives. (The report-only
    #   and enforcement headers may be used at the same time.)
    enforce_policies = getattr(settings, 'CSP_STATIC_ENFORCE', None)

    # .. setting_name: CSP_STATIC_REPORT_ONLY
    # .. setting_default: None
    # .. setting_description: Content-Security-Policy-Report-Only header to attach to
    #   all responses. See ``CSP_STATIC_ENFORCE`` for details.
    report_policies = getattr(settings, 'CSP_STATIC_REPORT_ONLY', None)

    # .. setting_name: CSP_STATIC_REPORTING_URI
    # .. setting_default: None
    # .. setting_description: URL of reporting server. This will be used for both Level 2 and
    #   Level 3 reports. If there are any semicolons or commas in the URL, they must be URL-encoded.
    #   Level 3 reporting is only enabled if ``CSP_STATIC_REPORTING_NAME`` is also set.
    reporting_uri = getattr(settings, 'CSP_STATIC_REPORTING_URI', None)

    # .. setting_name: CSP_STATIC_REPORTING_NAME
    # .. setting_default: None
    # .. setting_description: Used for CSP Level 3 reporting. This sets the name to use in the
    #   report-to CSP field and the Reporting-Endpoints header. If omitted, then Level 3 CSP
    #   reporting will not be enabled. If present, this must be a string starting with an ASCII
    #   letter and can contain ASCII letters, numbers, hyphen, underscore, and some other characters.
    #   See https://www.rfc-editor.org/rfc/rfc8941.html#section-3.3.4 for full grammar.
    reporting_endpoint_name = getattr(settings, 'CSP_STATIC_REPORTING_NAME', None)

    if not enforce_policies and not report_policies:
        return {}

    headers = {}

    reporting_suffix = ''
    if reporting_uri:
        reporting_suffix = f"; report-uri {reporting_uri}"
        if reporting_endpoint_name:
            headers['Reporting-Endpoints'] = f'{reporting_endpoint_name}="{reporting_uri}"'
            reporting_suffix += f"; report-to {reporting_endpoint_name}"

    def clean_header(value):
        # Collapse any internal whitespace that contains a newline. This allows
        # writing the setting value as a multi-line string, which is useful for
        # CSP -- the values can be quite long.
        value = re.sub("\\s*\n\\s*", " ", value).strip()
        # Remove any trailing semicolon, which we allow (for convenience).
        # The CSP spec does not allow trailing semicolons or empty directives.
        value = re.sub("[;\\s]+$", "", value)
        return value

    if enforce_policies:
        headers['Content-Security-Policy'] = clean_header(enforce_policies) + reporting_suffix

    if report_policies:
        headers['Content-Security-Policy-Report-Only'] = clean_header(report_policies) + reporting_suffix

    return headers


def _append_headers(response_headers, more_headers):
    """
    Append to the response headers. If a header already exists, assume it is
    permitted to be multi-valued (comma-separated), and update the existing value.

    Arguments:
        response_headers: response.headers (or any dict-like object), to be modified
        more_headers: Dict of header names to values
    """
    for k, v in more_headers.items():
        if existing := response_headers.get(k):
            response_headers[k] = f"{existing}, {v}"
        else:
            response_headers[k] = v


def content_security_policy_middleware(get_response):
    """
    Middleware that adds Content-Security-Policy and related headers, if enabled.

    This should be reasonably high up in the middleware chain, since it should
    apply to all responses.
    """
    # Constant across all requests, since they're based on static settings.
    csp_headers = _load_headers()
    if not csp_headers:
        raise MiddlewareNotUsed()  # tell Django to skip this middleware

    def middleware_handler(request):
        response = get_response(request)
        # Reporting-Endpoints, CSP, and CSP-RO can all be multi-valued
        # (comma-separated) headers, though the CSP spec says "SHOULD NOT"
        # for the latter two.
        _append_headers(response.headers, csp_headers)
        return response

    return middleware_handler



"""
This script takes a regex to search through the NRQL of New Relic alert policies
and New dashboards.

This script makes use of New Relic's GraphQL API. See https://api.newrelic.com/graphiql.

For help::

    python edx_django_utils/monitoring/scripts/new_relic_nrql_search.py --help

"""
import json
import os
import re
from string import Template

import click
import requests


@click.command()
@click.option(
    '--regex',
    required=True,
    help="The regex to use to search NRQL in alert policies and dashboards.",
)
@click.option(
    '--policy_id',
    multiple=True,
    help="Optionally provide a specific policy id to check. Multiple can be supplied.",
)
@click.option(
    '--dashboard_guid',
    multiple=True,
    help="Optionally provide a specific dashboard guid to check. Multiple can be supplied.",
)
def main(regex, policy_id, dashboard_guid):
    """
    Search NRQL in New Relic alert policies and dashboards using regex.

    Example usage:

        new_relic_nrql_search.py --regex tnl

    Note: The search ignores case since NRQL is case insensitive.

    Pre-requisite, set the following environment variable (in a safe way):

        NEW_RELIC_API_KEY=XXXXXXX

    See https://docs.newrelic.com/docs/apis/intro-apis/new-relic-api-keys/#user-api-key for details
    on setting an API key.

    To skip alert policies or dashboards, just use a non-existent id, like --policy_id 0 or --dashboard_guid 0.

    """
    # Set environment variables
    api_key = os.environ['NEW_RELIC_API_KEY']
    headers = {
        "Content-Type": "application/json",
        "X-Api-Key": api_key,
    }

    compiled_regex = re.compile(regex)

    account_ids = get_account_ids(headers)
    for account_id in account_ids:
        search_alert_policies(compiled_regex, account_id, headers, policy_id)
    print()
    search_dashboards(compiled_regex, headers, dashboard_guid)
    print(flush=True)


def get_account_ids(headers):
    """
    Returns a list of the New Relic account ids to be searched.
    """
    response = requests.get(
        'https://api.newrelic.com/graphql',
        headers=headers,
        params={'query': """
            {
              actor {
                accounts {
                  id
                }
              }
            }
        """},
    )
    response.raise_for_status()  # could be an error response
    response_data = response.json()
    account_ids = [account['id'] for account in response_data['data']['actor']['accounts']]
    return account_ids


# Template with variables:
# - account_id
# - cursor; can be null or cursor string
ALERT_POLICY_LIST_TEMPLATE = Template("""
    {
      actor {
        account(id: ${account_id}) {
          alerts {
            policiesSearch(cursor: ${cursor}) {
              policies {
                id
                name
              }
              nextCursor
            }
          }
        }
      }
    }
""")

# Template with variables:
# - account_id
# - policy_id
NRQL_ALERT_CONDITIONS_TEMPLATE = Template("""
    {
      actor {
        account(id: ${account_id}) {
          alerts {
            nrqlConditionsSearch(searchCriteria: {policyId: ${policy_id}}) {
              nrqlConditions {
                nrql {
                  query
                }
                name
              }
            }
          }
        }
      }
    }
""")


def search_alert_policies(regex, account_id, headers, policy_id):
    """
    Searches New Relic alert policy NRQL using the regex argument.

    Arguments:
        regex (re.Pattern): compiled regex used to compare against NRQL
        account_id (int): the id of the New Relic account in which to search.
        headers (dict): headers required to make http requests to New Relic.
        policy_id (tuple): optional tuple of policy ids supplied from the command-line.
    """
    policies = []
    cursor = 'null'
    while True:
        response = requests.get(
            'https://api.newrelic.com/graphql',
            headers=headers,
            params={'query': ALERT_POLICY_LIST_TEMPLATE.substitute(
                account_id=account_id,
                cursor=cursor
            )},
        )
        response.raise_for_status()  # could be an error response
        response_data = response.json()
        query_results = response_data['data']['actor']['account']['alerts']['policiesSearch']
        policies += query_results['policies']
        if not query_results['nextCursor']:
            break
        cursor = f"\"{query_results['nextCursor']}\""
    # Note: policy_id is an optional tuple of policy ids supplied from the command-line.
    if policy_id:
        policies = [policy for policy in policies if policy['id'] in policy_id]
    print(f"Searching for regex {regex.pattern} in {len(policies)} alert policies in account {account_id}...")
    policy_ids_printed = {}
    for policy in policies:
        print('.', end='', flush=True)

        # get the NRQL alert conditions from the alert policy
        response = requests.get(
            'https://api.newrelic.com/graphql',
            headers=headers,
            params={'query': NRQL_ALERT_CONDITIONS_TEMPLATE.substitute(
                account_id=account_id,
                policy_id=policy['id'],
            )},
        )
        response.raise_for_status()  # could be an error response
        response_data = response.json()

        nrql_conditions = response_data['data']['actor']['account']['alerts']['nrqlConditionsSearch']['nrqlConditions']
        for nrql_condition in nrql_conditions:
            nrql_query = nrql_condition['nrql']['query']
            if regex.search(nrql_query, re.IGNORECASE):

                # Print the alert policy header for the first alert condition matched
                if policy['id'] not in policy_ids_printed:
                    policy_ids_printed[policy['id']] = True
                    print('\n')
                    print(
                        f"Found in \"{policy['name']}\" "
                        # NOTE: The API doesn't provide a link to the policy, so this is a static link to
                        #   the alert policies home page.
                        f"(policy_id={policy['id']}, search_link=https://one.nr/0X8woZOZvQx):"
                    )
                    print('')

                # Print the alert condition that matched
                print(f"- {nrql_condition['name']}: {nrql_query}")

    if policy_ids_printed:
        command_line = ''
        for policy_id in policy_ids_printed.keys():
            command_line += f'--policy_id {policy_id} '
        print("\n\nRun again with found policies: {}".format(command_line))
    else:
        print("\n\nNo alert policies matched.")


# Retrieves list of dashboard and dashboard page entities
# Template with variable cursor; can be null or cursor string
DASHBOARD_LIST_QUERY_TEMPLATE = Template("""
    {
      actor {
        entitySearch(queryBuilder: {type: DASHBOARD}) {
          results(cursor: ${cursor}) {
            entities {
              ... on DashboardEntityOutline {
                guid
                name
                accountId
                dashboardParentGuid
                permalink
              }
            }
            nextCursor
          }
          count
        }
      }
    }
""")

# Retrieves dashboard entities. Does not work for dashboard pages.
DASHBOARD_ENTITY_QUERY = """
    query ($guids: EntityGuid!) {
      actor {
        entities(guids: $guids) {
          ... on DashboardEntity {
            guid
            pages {
              widgets {
                rawConfiguration
                title
              }
            }
            name
          }
        }
      }
    }
"""


def search_dashboards(regex, headers, dashboard_guid):
    """
    Searches New Relic alert policy NRQL using the regex argument.

    Arguments:
        regex (re.Pattern): compiled regex used to compare against NRQL
        headers (dict): headers required to make http requests to New Relic
        dashboard_guid (tuple): optional tuple of dashboard guids supplied from the command-line.
    """
    # load details of all dashboards
    dashboards = []
    cursor = 'null'
    while True:
        response = requests.get(
            'https://api.newrelic.com/graphql',
            headers=headers,
            params={'query': DASHBOARD_LIST_QUERY_TEMPLATE.substitute(cursor=cursor)},
        )
        response.raise_for_status()  # could be an error response
        response_data = response.json()
        query_results = response_data['data']['actor']['entitySearch']['results']
        dashboards += query_results['entities']
        if not query_results['nextCursor']:
            break
        cursor = f"\"{query_results['nextCursor']}\""
    # Filter out dashboard pages, which the dashboard entity query will not be able to find.
    # - Note: This could probably be handled in the original query, but not sure how.
    dashboards = [dashboard for dashboard in dashboards if dashboard['dashboardParentGuid'] is None]
    # Note: dashboard_guid is an optional tuple of dashboard guids supplied from the command-line.
    if dashboard_guid:
        dashboards = [dashboard for dashboard in dashboards if dashboard['guid'] in dashboard_guid]
    print(f"Searching for regex {regex.pattern} in {len(dashboards)} dashboards...")
    dashboard_guids_printed = {}
    for dashboard in dashboards:
        print('.', end='', flush=True)

        # get the dashboard details
        response = requests.get(
            'https://api.newrelic.com/graphql',
            headers=headers,
            params={
                'query': DASHBOARD_ENTITY_QUERY,
                'variables': json.dumps({'guids': dashboard['guid']}),
            }
        )
        response.raise_for_status()  # could be an error response
        response_data = response.json()

        if response_data['data']['actor']['entities'][0]['pages']:
            for page in response_data['data']['actor']['entities'][0]['pages']:
                for widget in page['widgets']:
                    if 'nrqlQueries' not in widget['rawConfiguration']:
                        continue

                    for nrql_query in widget['rawConfiguration']['nrqlQueries']:
                        query = nrql_query['query']
                        if regex.search(query, re.IGNORECASE):

                            # Print the dashboard header for the first widget nrql that matches
                            if dashboard['guid'] not in dashboard_guids_printed:
                                dashboard_guids_printed[dashboard['guid']] = True
                                print('\n')
                                print(
                                    f"Found in \"{dashboard['name']}\" "
                                    f"(guid={dashboard['guid']}, link={dashboard['permalink']}):"
                                )
                                print('')

                            # Print the widget NRQL that matches
                            print(f"- {widget['title']}: {query}")

    if dashboard_guids_printed:
        command_line = ''
        for dashboard_guid in dashboard_guids_printed.keys():
            command_line += f'--dashboard_guid {dashboard_guid} '
        print("\n\nRun again with found dashboards: {}".format(command_line))
    else:
        print("\n\nNo dashboards found that match.")


if __name__ == "__main__":
    main()  # pylint: disable=no-value-for-parameter



"""
Tests monitoring middleware.

Note: CachedCustomMonitoringMiddleware is tested in ``test_custom_monitoring.py``.
"""
import re
from unittest.mock import Mock, call, patch

import ddt
from django.test import TestCase
from django.test.client import RequestFactory
from django.test.utils import override_settings
from waffle.testutils import override_switch

from edx_django_utils.cache import RequestCache
from edx_django_utils.monitoring import (
    CookieMonitoringMiddleware,
    DeploymentMonitoringMiddleware,
    MonitoringMemoryMiddleware
)


class TestMonitoringMemoryMiddleware(TestCase):
    """
    Tests for MonitoringMemoryMiddleware
    """
    @override_switch('edx_django_utils.monitoring.enable_memory_middleware', False)
    @patch('edx_django_utils.monitoring.internal.middleware.log')
    def test_memory_monitoring_when_disabled(self, mock_logger):
        MonitoringMemoryMiddleware().process_response(
            'fake request',
            'fake response',
        )
        mock_logger.info.assert_not_called()

    @override_switch('edx_django_utils.monitoring.enable_memory_middleware', True)
    @patch('edx_django_utils.monitoring.internal.middleware.log')
    def test_memory_monitoring_when_enabled(self, mock_logger):
        request = RequestFactory().get('/')
        MonitoringMemoryMiddleware().process_response(
            request,
            'fake response',
        )
        mock_logger.info.assert_called()


class TestDeploymentMonitoringMiddleware(TestCase):
    """
    Test the DeploymentMonitoringMiddleware functionalities
    """
    version_pattern = r'\d+(\.\d+){2}'

    def setUp(self):
        super().setUp()
        RequestCache.clear_all_namespaces()

    def _test_key_value_pair(self, function_call, key):
        """
        Asserts the function call key and value with the provided key and the default version_pattern
        """
        attribute_key, attribute_value = function_call[0]
        assert attribute_key == key
        assert re.match(re.compile(self.version_pattern), attribute_value)

    @patch('newrelic.agent')
    def test_record_python_and_django_version(self, mock_newrelic_agent):
        """
        Test that the DeploymentMonitoringMiddleware records the correct Python and Django versions
        """
        middleware = DeploymentMonitoringMiddleware(Mock())
        middleware(Mock())

        parameter_calls_count = mock_newrelic_agent.add_custom_parameter.call_count
        assert parameter_calls_count == 2

        function_calls = mock_newrelic_agent.add_custom_parameter.call_args_list
        self._test_key_value_pair(function_calls[0], 'python_version')
        self._test_key_value_pair(function_calls[1], 'django_version')


@ddt.ddt
class CookieMonitoringMiddlewareTestCase(TestCase):
    """
    Tests for CookieMonitoringMiddleware.
    """
    def setUp(self):
        super().setUp()
        self.mock_response = Mock()

    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch("edx_django_utils.monitoring.internal.middleware._set_custom_attribute")
    @ddt.data(
        (None, None),  # logging threshold not defined
        (5, None),  # logging threshold too high
        (5, 9999999999999999999),  # logging threshold too high, and random sampling impossibly unlikely
    )
    @ddt.unpack
    def test_cookie_monitoring_with_no_logging(
        self, logging_threshold, sampling_request_count, mock_set_custom_attribute, mock_logger
    ):
        expected_response = self.mock_response
        middleware = CookieMonitoringMiddleware(lambda request: expected_response)
        cookies_dict = {'a': 'y'}

        with override_settings(COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=logging_threshold):
            with override_settings(COOKIE_SAMPLING_REQUEST_COUNT=sampling_request_count):
                actual_response = middleware(self.get_mock_request(cookies_dict))

        assert actual_response == expected_response
        # expect monitoring of header size for all requests
        mock_set_custom_attribute.assert_called_once_with('cookies.header.size', 3)
        # cookie logging was not enabled, so nothing should be logged
        mock_logger.info.assert_not_called()
        mock_logger.exception.assert_not_called()

    @override_settings(COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=None)
    @override_settings(COOKIE_SAMPLING_REQUEST_COUNT=None)
    @patch("edx_django_utils.monitoring.internal.middleware._set_custom_attribute")
    @ddt.data(
        # A corrupt cookie header contains "Cookie: ".
        ('corruptCookie: normal-cookie=value', 1, 1),
        ('corrupt1Cookie: normal-cookie1=value1;corrupt2Cookie: normal-cookie2=value2', 2, 2),
        ('corrupt=Cookie: value', 1, 0),
    )
    @ddt.unpack
    def test_cookie_header_corrupt_monitoring(
        self, corrupt_cookie_header, expected_corrupt_count, expected_corrupt_key_count, mock_set_custom_attribute
    ):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        request = RequestFactory().request()
        request.META['HTTP_COOKIE'] = corrupt_cookie_header

        middleware(request)

        mock_set_custom_attribute.assert_has_calls([
            call('cookies.header.size', len(request.META['HTTP_COOKIE'])),
            call('cookies.header.corrupt_count', expected_corrupt_count),
            call('cookies.header.corrupt_key_count', expected_corrupt_key_count),
        ])

    @override_settings(COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=1)
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch("edx_django_utils.monitoring.internal.middleware._set_custom_attribute")
    def test_log_cookie_with_threshold_met(self, mock_set_custom_attribute, mock_logger):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        cookies_dict = {
            "a": "yy",
            "b": "xxx",
            "c": "z",
        }

        middleware(self.get_mock_request(cookies_dict))

        mock_set_custom_attribute.assert_has_calls([
            call('cookies.header.size', 16),
            call('cookies.header.size.computed', 16)
        ])
        mock_logger.info.assert_called_once_with(
            "Large (>= 1) cookie header detected. BEGIN-COOKIE-SIZES(total=16) b: 3, a: 2, c: 1 END-COOKIE-SIZES"
        )
        mock_logger.exception.assert_not_called()

    @override_settings(
        COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=1,
        UNUSUAL_COOKIE_HEADER_PUBLIC_KEY="some private key",
    )
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch("edx_django_utils.monitoring.internal.middleware.encrypt_for_log", return_value="[encrypted: 50M3JUN|<]")
    def test_log_corrupt_cookies(self, mock_encrypt, mock_logger):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        request = RequestFactory().request()
        request.META['HTTP_COOKIE'] = 'aa=1; bCookie: bb=22; ccc=3Cookie: 33'
        request.META['HTTP_SOMETHING'] = 'else'

        middleware(request)

        # Is passed all headers
        mock_encrypt.assert_called_once_with(
            '{"Cookie": "aa=1; bCookie: bb=22; ccc=3Cookie: 33", "Something": "else"}',
            "some private key",
        )

        # Encrypted headers are logged in additional message if any cookies appeared corrupted.
        mock_logger.info.assert_has_calls([
            call("All headers for request with corrupted cookies (count=2): [encrypted: 50M3JUN|<]"),
            call(
                "Large (>= 1) cookie header detected. "
                "BEGIN-COOKIE-SIZES(total=37) ccc: 11, bCookie: bb: 2, aa: 1 END-COOKIE-SIZES"
            ),
        ])

    @override_settings(
        COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=1,
        UNUSUAL_COOKIE_HEADER_PUBLIC_KEY="some private key",
        UNUSUAL_COOKIE_HEADER_LOG_CHUNK=75,
    )
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch(
        "edx_django_utils.monitoring.internal.middleware.encrypt_for_log",
        return_value="[encrypted: 50M3JUN|<_aaaabbbbccccdddd]"
    )
    def test_log_corrupt_cookies_split(self, mock_encrypt, mock_logger):
        """
        Like ``test_log_corrupt_cookies`` but with message splitting.
        """
        middleware = CookieMonitoringMiddleware(self.mock_response)
        request = RequestFactory().request()
        request.META['HTTP_COOKIE'] = 'aa=1; bCookie: bb=22; ccc=3Cookie: 33'
        request.META['HTTP_SOMETHING'] = 'else'

        middleware(request)

        # Is passed all headers
        mock_encrypt.assert_called_once_with(
            '{"Cookie": "aa=1; bCookie: bb=22; ccc=3Cookie: 33", "Something": "else"}',
            "some private key",
        )

        # Encrypted headers are logged across multiple messages if too large.
        mock_logger.info.assert_has_calls([
            call(
                "All headers for request with corrupted cookies (count=2): "
                "[encrypted: 50M3J [chunk #1, group=OUMTkx5O continues]"
            ),
            call("UN|<_aaaabbbbccccdddd] [chunk #2, group=OUMTkx5O final]"),
            call(
                "Large (>= 1) cookie header detected. "
                "BEGIN-COOKIE-SIZES(total=37) ccc: 11, bCookie: bb: 2, aa: 1 END-COOKIE-SIZES"
            ),
        ])

    @override_settings(COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=9999)
    @override_settings(COOKIE_SAMPLING_REQUEST_COUNT=1)
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch("edx_django_utils.monitoring.internal.middleware._set_custom_attribute")
    def test_log_cookie_with_sampling(self, mock_set_custom_attribute, mock_logger):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        cookies_dict = {
            "a": "yy",
            "b": "xxx",
            "c": "z",
        }

        middleware(self.get_mock_request(cookies_dict))

        mock_set_custom_attribute.assert_has_calls([
            call('cookies.header.size', 16),
            call('cookies.header.size.computed', 16)
        ])
        mock_logger.info.assert_called_once_with(
            "Sampled small (< 9999) cookie header. BEGIN-COOKIE-SIZES(total=16) b: 3, a: 2, c: 1 END-COOKIE-SIZES"
        )
        mock_logger.exception.assert_not_called()

    @override_settings(COOKIE_HEADER_SIZE_LOGGING_THRESHOLD=9999)
    @override_settings(COOKIE_SAMPLING_REQUEST_COUNT=1)
    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    @patch("edx_django_utils.monitoring.internal.middleware._set_custom_attribute")
    def test_empty_cookie_header_skips_sampling(self, mock_set_custom_attribute, mock_logger):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        cookies_dict = {}

        middleware(self.get_mock_request(cookies_dict))

        mock_set_custom_attribute.assert_has_calls([
            call('cookies.header.size', 0),
        ])
        mock_logger.info.assert_not_called()
        mock_logger.exception.assert_not_called()

    @patch('edx_django_utils.monitoring.internal.middleware.log', autospec=True)
    def test_cookie_monitoring_unknown_exception(self, mock_logger):
        middleware = CookieMonitoringMiddleware(self.mock_response)
        cookies_dict = {'a': 'y'}
        mock_request = self.get_mock_request(cookies_dict)
        mock_request.META = Mock()
        mock_request.META.side_effect = Exception("Some exception")

        middleware(mock_request)

        mock_logger.exception.assert_called_once_with("Unexpected error logging and monitoring cookies.")

    def get_mock_request(self, cookies_dict):
        """
        Return mock request with the provided cookies in the header.
        """
        factory = RequestFactory()
        for name, value in cookies_dict.items():
            factory.cookies[name] = value
        return factory.request()



"""
Code to create Registry of django app plugins

Please remember to expose any new public methods in the `__init__.py` file.
"""
from .plugin_manager import PluginManager


class DjangoAppRegistry(PluginManager):
    """
    DjangoAppRegistry is a registry of django app plugins.
    """


def get_plugin_app_configs(project_type):
    return DjangoAppRegistry.get_available_plugins(project_type).values()



"""
Management command `manage_user` is used to idempotently create or remove
Django users, set/unset permission bits, and associate groups by name.
"""


from django.contrib.auth import get_user_model
from django.contrib.auth.hashers import identify_hasher, is_password_usable
from django.contrib.auth.models import Group
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction
from django.utils.translation import gettext as _

from edx_django_utils.user import generate_password


def is_valid_django_hash(encoded):
    """
    Starting with django 2.1, the function is_password_usable no longer checks whether encode
    is a valid password created by a django hasher (hasher in  PASSWORD_HASHERS setting)

    Adding this function to create constant behavior as we upgrade django versions
    """
    try:
        identify_hasher(encoded)
    except ValueError:
        return False
    return True


class Command(BaseCommand):
    """Creates the specified user, if it does not exist, and sets its groups."""
    help = __doc__

    def add_arguments(self, parser):
        parser.add_argument('username')
        parser.add_argument('email')
        parser.add_argument('--remove', dest='is_remove', action='store_true')
        parser.add_argument('--superuser', dest='is_superuser', action='store_true')
        parser.add_argument('--staff', dest='is_staff', action='store_true')
        parser.add_argument('--unusable-password', dest='unusable_password', action='store_true')
        parser.add_argument('--initial-password-hash', dest='initial_password_hash')
        parser.add_argument('-g', '--groups', nargs='*', default=[])

    def _maybe_update(self, user, attribute, new_value):
        """
        DRY helper.  If the specified attribute of the user differs from the
        specified value, it will be updated.
        """
        old_value = getattr(user, attribute)
        if new_value != old_value:
            self.stderr.write(
                _('Setting {attribute} for user "{username}" to "{new_value}"').format(
                    attribute=attribute, username=user.username, new_value=new_value
                )
            )
            setattr(user, attribute, new_value)

    def _check_email_match(self, user, email):
        """
        DRY helper.

        Requiring the user to specify both username and email will help catch
        certain issues, for example if the expected username has already been
        taken by someone else.
        """
        if user.email.lower() != email.lower():
            # The passed email address doesn't match this username's email address.
            # Assume a problem and fail.
            raise CommandError(
                _(
                    'Skipping user "{}" because the specified and existing email '
                    'addresses "{}" and "{}" do not match.'
                ).format(user.username, user.email, email)
            )

    def _handle_remove(self, username, email):
        """
        If a user with the given username and email exists, delete them.
        """
        try:
            user = get_user_model().objects.get(username=username)
        except get_user_model().DoesNotExist:
            self.stderr.write(_('Did not find a user with username "{}" - skipping.').format(username))
            return
        self._check_email_match(user, email)
        self.stderr.write(_('Removing user: "{}"').format(user))
        user.delete()

    @transaction.atomic
    def handle(  # pylint: disable=arguments-differ
        self, username, email, is_remove, is_staff, is_superuser, groups,
        unusable_password, initial_password_hash, *args, **options
    ):

        if is_remove:
            self._handle_remove(username, email)
            return

        old_groups, new_groups = set(), set()
        user, created = get_user_model().objects.get_or_create(
            username=username,
            defaults={'email': email}
        )

        if created:
            if initial_password_hash:
                if not (is_password_usable(initial_password_hash) and is_valid_django_hash(initial_password_hash)):
                    raise CommandError(f'The password hash provided for user {username} is invalid.')
                user.password = initial_password_hash
            else:
                # Set the password to a random, unknown, but usable password
                # allowing self-service password resetting.  Cases where unusable
                # passwords are required, should be explicit, and will be handled below.
                user.set_password(generate_password(length=25))
            self.stderr.write(_('Created new user: "{}"').format(user))
        else:
            # NOTE, we will not update the email address of an existing user.
            self.stderr.write(_('Found existing user: "{}"').format(user))
            self._check_email_match(user, email)
            old_groups = set(user.groups.all())

        self._maybe_update(user, 'is_staff', is_staff)
        self._maybe_update(user, 'is_superuser', is_superuser)

        # Set unusable password if specified
        if unusable_password and user.has_usable_password():
            self.stderr.write(_('Setting unusable password for user "{}"').format(user))
            user.set_unusable_password()

        # resolve the specified groups
        for group_name in groups or set():

            try:
                group = Group.objects.get(name=group_name)
                new_groups.add(group)
            except Group.DoesNotExist:
                # warn, but move on.
                self.stderr.write(_('Could not find a group named "{}" - skipping.').format(group_name))

        add_groups = new_groups - old_groups
        remove_groups = old_groups - new_groups

        self.stderr.write(
            _(
                'Adding user "{username}" to groups {group_names}'
            ).format(
                username=user.username,
                group_names=[g.name for g in add_groups]
            )
        )
        self.stderr.write(
            _(
                'Removing user "{username}" from groups {group_names}'
            ).format(
                username=user.username,
                group_names=[g.name for g in remove_groups]
            )
        )

        user.groups.set(new_groups)
        user._called_by_management_command = True  # pylint: disable=protected-access
        user.save()



"""
Public API for determining the IP address of a request.

Summary
=======

For developers:

- Call ``get_safest_client_ip`` whenever you want to know the caller's IP address
- Make sure ``init_client_ips`` is called as early as possible in the middleware stack
- See the "Guidance for developers" section for more advanced usage

For site operators:

- See the "Configuration" section for important information and guidance

For everyone:

- Background information is available in the "Concepts" section


Concepts
========

- The *IP chain* is the list of IPs in the ``X-Forwarded-For`` (XFF) header followed
  by the ``REMOTE_ADDR`` value. If all involved parties are telling the truth, this
  is the list of IP addresses that have relayed the HTTP request. However, due to
  the possibility of spoofing, this raw data cannot be used directly for all
  purposes:

  - The rightmost IP in the chain is the IP that has directly connected with the
    server and sent or relayed the request. In most deployments, this is likely
    to be a reverse proxy such as nginx. In any case it is the "closest" IP (in
    the sense of the request chain, not in terms of geographic proximity.)
  - The next closest IP, if present, is the one that the closest IP *claims*
    sent the request to it. Each IP in the chain can only vouch for the
    correctness of the IP immediately to its left in the list.
  - In a normal, unspoofed request, the leftmost IP is the "real" client IP, the
    IP of the computer that made the original request.
  - However, clients can send a fake XFF header, so the leftmost IP in the chain
    cannot be trusted in the general case. In fact, the only IP that can be
    trusted absolutely is the rightmost one.
  - The challenge is to determine what the leftmost *trusted* IP is, as this is
    the most accurate we can get without compromising on security.

- The *external chain* is some prefix of the IP chain that stops before the
  (recognized) edge of the deployment's infrastructure. That is, the external
  chain is the portion of the IP chain that is to the left of some trust
  boundary, as determined by configuration or some fallback method. This is the
  list of IPs that can all plausibly be considered the "real" IP of the client.
  If the server is configured correctly this may contain, in order: Any IPs
  spoofed by the client, the client's own IP, IPs of any forwarding HTTP proxies
  specified by the client, and then IPs of any reverse HTTP proxies the
  request passed through *before* reaching the deployment's own infrastructure
  (CDN, load balancer, etc.)

  - Caveat: In the case where the request is being sent through an anonymizing
    proxy such as a VPN, the VPN's exit node IP is considered the "real" client
    IP.
  - Despite the name, this chain may contain private-range IP addresses, in
    particular if a request originates from another server in the same
    datacenter.


Guidance for developers
=======================

Almost anywhere you care about IP address, just call ``get_safest_client_ip``.
This will get you the *rightmost* IP of the external chain (defined above).
Because it cannot be easily spoofed by the caller, it is suitable for adversarial
use-cases such as:

- Rate-limiting
- Only allowing certain IPs to access a resource (or alternatively, blocking them)

In some less common situations where you need the entire external chain, you
can call ``get_all_client_ips`. This returns a list of IP addresses, although for
the great majority of normal requests this will be a list of length 1. This list is
appropriate for when you're recording IPs for manual review or need to make a
decision based on all of the IPs (no matter which one is the "real" one. This might
include:

- Audit logs
- Telling a user about other active sessions on their account
- Georestriction

In some very rare cases you might want just a single IP that isn't rightmost. In
some cases you might ask for the entire external chain and then take the leftmost
IP. This should only be used in non-adversarial situations, and is usually the wrong
choice, but may be appropriate for:

- Localization (if other HTTP headers aren't sufficient)
- Analytics


Configuration
=============

Configuration is via ``CLOSEST_CLIENT_IP_FROM_HEADERS``, which allows specifying
an HTTP header that will be trusted to report the rightmost IP in the external chain.
See setting annotation for details, but guidance on common configurations is provided
here:

- If you use a CDN as your outermost proxy:

  - Find what header your CDN sends to its origin that indicates the remote address it
    sees on inbound connections. For example, with Cloudflare this is ``CF-Connecting-IP``.
  - Ensure that your CDN always overrides this header if it exists in the inbound request,
    and never accepts a value provided by the client. Some CDNs are better than others
    about this.
  - Recommended setting, using Cloudflare as the example::

       CLOSEST_CLIENT_IP_FROM_HEADERS:
       - name: CF-Connecting-IP
         index: 0

    It would be equivalent to use ``-1`` as the index since there is always one and only
    one IP in this header, and Python list indexing rules are used here.
  - As a general rule, you should also ensure that traffic cannot bypass the CDN and reach
    your origin directly, since otherwise attackers will be able to spoof their IP address
    (and bypass protections your CDN provides). You may need to arrange for your CDN to set
    a header containing a shared secret.

- If your outermost proxy is an AWS ELB or other proxy on the same local network as your
  server, or you have any other configuration in which your proxies and application speak
  to each other using private-range IP addresses:

    - You can rely on the rightmost public IP in the IP chain to be the safest client IP.
      To do this, set your configuration for zero trusted headers::

         CLOSEST_CLIENT_IP_FROM_HEADERS: []

    - This assumes that 1) your outermost proxy always appends to ``X-Forwarded-For``, and
      2) any further proxies between that one and your application either append to it
      (ideal) or pass it along unchanged (not ideal, but workable). This is true by default
      for most proxy software.

- If you have any reverse proxy that will be seen by the next proxy or your application as
  having a public IP:

  - You'll need to rely on having a consistent *number* of proxies in front of your
    application, and you'll need to know which ones append to the ``X-Forwarded-For``
    header instead of just passing it unchanged.
  - Once you know the number of your proxies in the chain that append, you can use this
    count to say that the Nth-from-last IP in the ``X-Forwarded-For`` is the closest client
    IP. For example, if you had two, you would use ``-2`` (note the negative sign) to
    indicate the second-from-last IP::

       CLOSEST_CLIENT_IP_FROM_HEADERS:
       - name: X-Forwarded-For
         index: -2

  - This is fragile in the face of network configuration changes, so having your outermost
    proxy set a special header is preferred.
  - Configuring the proxy count too low will result in rate-limiting your own proxies;
    configuring it too high will allow attackers to bypass rate-limiting.
  - Side note: Even if you don't use it for ``CLOSEST_CLIENT_IP_FROM_HEADERS``, this
    proxy-counting approach will be required for configuring django-rest-framework's
    ``NUM_PROXIES`` setting.

- If your application is directly exposed to the public internet, without even a local proxy:

  - This is an unusual configuration, but simple to configure; with no proxies, just indicate
    that there are no trusted headers and therefore the closest public IP should be used::

       CLOSEST_CLIENT_IP_FROM_HEADERS: []
"""

from .internal.ip import get_all_client_ips, get_raw_ip_chain, get_safest_client_ip, init_client_ips



"""
Middleware for monitoring.

At this time, monitoring details can only be reported to New Relic.

"""
import base64
import hashlib
import json
import logging
import math
import platform
import random
import warnings
from uuid import uuid4

import django
import psutil
import waffle  # pylint: disable=invalid-django-waffle-import
from django.conf import settings
from django.utils.deprecation import MiddlewareMixin

from edx_django_utils.cache import RequestCache
from edx_django_utils.logging import encrypt_for_log

log = logging.getLogger(__name__)
try:
    import newrelic.agent
except ImportError:  # pragma: no cover
    log.warning("Unable to load NewRelic agent module")
    newrelic = None  # pylint: disable=invalid-name


_DEFAULT_NAMESPACE = 'edx_django_utils.monitoring'
_REQUEST_CACHE_NAMESPACE = f'{_DEFAULT_NAMESPACE}.custom_attributes'


class DeploymentMonitoringMiddleware:
    """
    Middleware to record environment values at the time of deployment for each service.
    """
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        self.record_python_version()
        self.record_django_version()
        response = self.get_response(request)
        return response

    @staticmethod
    def record_django_version():
        """
        Record the installed Django version as custom attribute

        .. custom_attribute_name: django_version
        .. custom_attribute_description: The django version in use (e.g. '2.2.24').
           Set by DeploymentMonitoringMiddleware.
        """
        _set_custom_attribute('django_version', django.__version__)

    @staticmethod
    def record_python_version():
        """
        Record the Python version as custom attribute

        .. custom_attribute_name: python_version
        .. custom_attribute_description: The Python version in use (e.g. '3.8.10').
           Set by DeploymentMonitoringMiddleware.
        """
        _set_custom_attribute('python_version', platform.python_version())


class CachedCustomMonitoringMiddleware(MiddlewareMixin):
    """
    Middleware batch reports cached custom attributes at the end of a request.

    Make sure to add below the request cache in MIDDLEWARE.

    This middleware will only call on the newrelic agent if there are any attributes
    to report for this request, so it will not incur any processing overhead for
    request handlers which do not record custom attributes.

    Note: New Relic adds custom attributes to events, which is what is being used here.

    """
    @classmethod
    def _get_attributes_cache(cls):
        """
        Get a request cache specifically for New Relic custom attributes.
        """
        return RequestCache(namespace=_REQUEST_CACHE_NAMESPACE)

    @classmethod
    def accumulate_attribute(cls, name, value):
        """
        Accumulate a custom attribute (name and value) in the attributes cache.
        """
        attributes_cache = cls._get_attributes_cache()
        cached_response = attributes_cache.get_cached_response(name)
        if cached_response.is_found:
            try:
                accumulated_value = value + cached_response.value
            except TypeError:
                _set_custom_attribute(
                    'error_adding_accumulated_metric',
                    'name={}, new_value={}, cached_value={}'.format(
                        name, repr(value), repr(cached_response.value)
                    )
                )
                return
        else:
            accumulated_value = value
        attributes_cache.set(name, accumulated_value)

    @classmethod
    def accumulate_metric(cls, name, value):  # pragma: no cover
        """
        Deprecated method replaced by accumulate_attribute.
        """
        msg = "Use 'accumulate_attribute' in place of 'accumulate_metric'."
        warnings.warn(msg, DeprecationWarning)
        _set_custom_attribute('deprecated_accumulate_metric', True)
        cls.accumulate_attribute(name, value)

    @classmethod
    def _batch_report(cls):
        """
        Report the collected custom attributes to New Relic.
        """
        if not newrelic:  # pragma: no cover
            return
        attributes_cache = cls._get_attributes_cache()
        for key, value in attributes_cache.data.items():
            _set_custom_attribute(key, value)

    # Whether or not there was an exception, report any custom NR attributes that
    # may have been collected.

    def process_response(self, request, response):
        """
        Django middleware handler to process a response
        """
        self._batch_report()
        return response

    def process_exception(self, request, exception):    # pylint: disable=W0613
        """
        Django middleware handler to process an exception
        """
        self._batch_report()


def _set_custom_attribute(key, value):
    """
    Sets monitoring custom attribute.

    Note: Can't use public method in ``utils.py`` due to circular reference.
    """
    if newrelic:  # pragma: no cover
        newrelic.agent.add_custom_parameter(key, value)


class MonitoringMemoryMiddleware(MiddlewareMixin):
    """
    Middleware for monitoring memory usage.

    Make sure to add below the request cache in MIDDLEWARE.
    """
    memory_data_key = 'memory_data'
    guid_key = 'guid_key'

    def process_request(self, request):
        """
        Store memory data to log later.
        """
        if self._is_enabled():
            self._cache.set(self.guid_key, str(uuid4()))
            log_prefix = self._log_prefix("Before", request)
            self._cache.set(self.memory_data_key, self._memory_data(log_prefix))

    def process_response(self, request, response):
        """
        Logs memory data after processing response.
        """
        if self._is_enabled():
            log_prefix = self._log_prefix("After", request)
            new_memory_data = self._memory_data(log_prefix)

            log_prefix = self._log_prefix("Diff", request)
            cached_memory_data_response = self._cache.get_cached_response(self.memory_data_key)
            old_memory_data = cached_memory_data_response.get_value_or_default(None)
            self._log_diff_memory_data(log_prefix, new_memory_data, old_memory_data)
        return response

    @property
    def _cache(self):
        """
        Namespaced request cache for tracking memory usage.
        """
        return RequestCache(namespace='monitoring_memory')

    def _log_prefix(self, prefix, request):
        """
        Returns a formatted prefix for logging for the given request.
        """
        # Note: After a celery task runs, the request cache is cleared. So if
        #   celery tasks are running synchronously (CELERY_ALWAYS _EAGER),
        #   "guid_key" will no longer be in the request cache when
        #   process_response executes.
        cached_guid_response = self._cache.get_cached_response(self.guid_key)
        cached_guid = cached_guid_response.get_value_or_default("without_guid")
        return f"{prefix} request '{request.method} {request.path} {cached_guid}'"

    def _memory_data(self, log_prefix):
        """
        Returns a dict with information for current memory utilization.
        Uses log_prefix in log statements.
        """
        machine_data = psutil.virtual_memory()

        process = psutil.Process()
        process_data = {
            'memory_info': process.memory_info(),
            'ext_memory_info': process.memory_info(),
            'memory_percent': process.memory_percent(),
            'cpu_percent': process.cpu_percent(),
        }

        log.info("%s Machine memory usage: %s; Process memory usage: %s", log_prefix, machine_data, process_data)
        return {
            'machine_data': machine_data,
            'process_data': process_data,
        }

    def _log_diff_memory_data(self, prefix, new_memory_data, old_memory_data):
        """
        Computes and logs the difference in memory utilization
        between the given old and new memory data.
        """
        def _vmem_used(memory_data):
            return memory_data['machine_data'].used

        def _process_mem_percent(memory_data):
            return memory_data['process_data']['memory_percent']

        def _process_rss(memory_data):
            return memory_data['process_data']['memory_info'].rss

        def _process_vms(memory_data):
            return memory_data['process_data']['memory_info'].vms

        if new_memory_data and old_memory_data:
            log.info(
                "%s Diff Vmem used: %s, Diff percent memory: %s, Diff rss: %s, Diff vms: %s",
                prefix,
                _vmem_used(new_memory_data) - _vmem_used(old_memory_data),
                _process_mem_percent(new_memory_data) - _process_mem_percent(old_memory_data),
                _process_rss(new_memory_data) - _process_rss(old_memory_data),
                _process_vms(new_memory_data) - _process_vms(old_memory_data),
            )

    def _is_enabled(self):
        """
        Returns whether this middleware is enabled.
        """
        return waffle.switch_is_active('edx_django_utils.monitoring.enable_memory_middleware')


class CookieMonitoringMiddleware:
    """
    Middleware for monitoring the size and growth of all our cookies, to see if
    we're running into browser limits.
    """
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Monitor at request-time to skip any cookies that may be added during the request.
        log_message = None
        try:
            log_message = self.get_log_message_and_monitor_cookies(request)
        except BaseException:
            log.exception("Unexpected error logging and monitoring cookies.")

        response = self.get_response(request)

        # Delay logging until response-time so that the user id can be included in the log message.
        if log_message:
            log.info(log_message)

        return response

    def get_log_message_and_monitor_cookies(self, request):
        """
        Add logging and custom attributes for monitoring cookie sizes.

        For cookie size monitoring, we don't log raw contents of cookies because that might
        cause a security issue—we just want to see if any cookies are growing out of control.
        However, there is also an option for encrypted logging of cookie data where there
        appears to be some kind of data corruption occurring.

        Useful NRQL Queries:

            # Always available
            SELECT * FROM Transaction WHERE cookies.header.size > 6000

        Attributes that are added by this middleware:

            For all requests:

                cookies.header.size: The total size in bytes of the cookie header

            If COOKIE_HEADER_SIZE_LOGGING_THRESHOLD is reached:

                cookies.header.size.computed

        Related Settings (see annotations for details):

            - COOKIE_HEADER_SIZE_LOGGING_THRESHOLD
            - COOKIE_SAMPLING_REQUEST_COUNT
            - UNUSUAL_COOKIE_HEADER_PUBLIC_KEY
            - UNUSUAL_COOKIE_HEADER_LOG_CHUNK

        Returns: The message to be logged. This is returned, rather than directly
            logged, so that it can be processed at request time (before any cookies
            may be changed server-side), but logged at response time, once the user
            id is available for authenticated calls.

        """
        raw_header_cookie = request.META.get('HTTP_COOKIE', '')
        cookie_header_size = len(raw_header_cookie.encode('utf-8'))
        # .. custom_attribute_name: cookies.header.size
        # .. custom_attribute_description: The total size in bytes of the cookie header.
        _set_custom_attribute('cookies.header.size', cookie_header_size)

        if cookie_header_size == 0:
            return None

        if corrupt_cookie_count := raw_header_cookie.count('Cookie: '):
            # .. custom_attribute_name: cookies.header.corrupt_count
            # .. custom_attribute_description: The attribute will only appear for potentially corrupt cookie headers,
            #   where "Cookie: " is found in the header. If this custom attribute is seen on the same
            #   requests where other mysterious cookie problems are occurring, this may help troubleshoot.
            #   See https://openedx.atlassian.net/browse/CR-4614 for more details.
            #   Also see cookies.header.corrupt_key_count
            _set_custom_attribute('cookies.header.corrupt_count', corrupt_cookie_count)
            # .. custom_attribute_name: cookies.header.corrupt_key_count
            # .. custom_attribute_description: The attribute will only appear for potentially corrupt cookie headers,
            #   where "Cookie: " is found in some of the cookie keys. If this custom attribute is seen on the same
            #   requests where other mysterious cookie problems are occurring, this may help troubleshoot.
            #   See https://openedx.atlassian.net/browse/CR-4614 for more details.
            #   Also see cookies.header.corrupt_count.
            _set_custom_attribute(
                'cookies.header.corrupt_key_count',
                sum(1 for key in request.COOKIES.keys() if 'Cookie: ' in key)
            )
            # If we have indication of corruption, just log all the headers for later diagnosis.
            # (Not part of other cookie logging because we need as much space as possible for
            # this log message, which can be quite large, and may need to chunk it across
            # multiple lines.)
            self.log_corrupt_cookie_headers(request, corrupt_cookie_count)

        # .. setting_name: COOKIE_HEADER_SIZE_LOGGING_THRESHOLD
        # .. setting_default: None
        # .. setting_description: The minimum size for the full cookie header to log a list of cookie names and sizes.
        #   Should be set to a relatively high threshold (suggested 9-10K) to avoid flooding the logs.
        logging_threshold = getattr(settings, "COOKIE_HEADER_SIZE_LOGGING_THRESHOLD", None)

        if not logging_threshold:
            return None

        is_large_cookie_header_detected = cookie_header_size >= logging_threshold
        if not is_large_cookie_header_detected:
            # .. setting_name: COOKIE_SAMPLING_REQUEST_COUNT
            # .. setting_default: None
            # .. setting_description: This setting enables sampling cookie header logging for cookie headers smaller
            #   than COOKIE_HEADER_SIZE_LOGGING_THRESHOLD. The cookie header logging will happen randomly for each
            #   request with a chance of 1 in COOKIE_SAMPLING_REQUEST_COUNT. For example, to see approximately one
            #   sampled log message every 10 minutes, set COOKIE_SAMPLING_REQUEST_COUNT to the average number of
            #   requests in 10 minutes.
            # .. setting_warning: This setting requires COOKIE_HEADER_SIZE_LOGGING_THRESHOLD to be enabled to take
            #   effect.
            sampling_request_count = getattr(settings, "COOKIE_SAMPLING_REQUEST_COUNT", None)

            # if the cookie header size is lower than the threshold, skip logging unless configured to do
            #   random sampling and we choose the lucky number (in this case, 1).
            if not sampling_request_count or random.randint(1, sampling_request_count) > 1:
                return None

        # The computed header size can be used to double check that there aren't large cookies that are
        #   duplicates in the original header (from different domains) that aren't being accounted for.
        cookies_header_size_computed = max(
            0, sum(len(name) + len(value) + 3 for (name, value) in request.COOKIES.items()) - 2
        )

        # .. custom_attribute_name: cookies.header.size.computed
        # .. custom_attribute_description: The computed total size in bytes of the cookie header, based on the
        #   cookies found in request.COOKIES. This value will only be captured for cookie headers larger than
        #   COOKIE_HEADER_SIZE_LOGGING_THRESHOLD. The value can be used to double check that there aren't large
        #   cookies that are duplicates in the cookie header (from different domains) that aren't being accounted
        #   for.
        _set_custom_attribute('cookies.header.size.computed', cookies_header_size_computed)

        # Sort starting with largest cookies
        sorted_cookie_items = sorted(request.COOKIES.items(), key=lambda x: len(x[1]), reverse=True)
        sizes = ', '.join(f"{name}: {len(value)}" for (name, value) in sorted_cookie_items)
        if is_large_cookie_header_detected:
            log_prefix = f"Large (>= {logging_threshold}) cookie header detected."
        else:
            log_prefix = f"Sampled small (< {logging_threshold}) cookie header."
        log_message = f"{log_prefix} BEGIN-COOKIE-SIZES(total={cookie_header_size}) {sizes} END-COOKIE-SIZES"

        return log_message

    def log_corrupt_cookie_headers(self, request, corrupt_cookie_count):
        """
        Log all headers when corrupt cookies are detected (if settings permit).

        This log data is encrypted using the log-sensitive utility.

        - Logging requires that ``UNUSUAL_COOKIE_HEADER_PUBLIC_KEY`` is set.
        - Output is split across multiple lines using ``UNUSUAL_COOKIE_HEADER_LOG_CHUNK``.
        """
        # Caller should ensure this, but check here anyway.
        if corrupt_cookie_count < 1:
            return

        # .. setting_name: UNUSUAL_COOKIE_HEADER_PUBLIC_KEY
        # .. setting_default: None
        # .. setting_description: Use this public key to encrypt and log headers when there's
        #   a corrupted-looking cookie in the request. See log_sensitive module for more detail
        #   on the encryption. If no key is provided, this logging is skipped. Also see
        #   ``UNUSUAL_COOKIE_HEADER_LOG_CHUNK``.
        corrupt_cookie_log_pub_key = getattr(settings, 'UNUSUAL_COOKIE_HEADER_PUBLIC_KEY', None)

        if not corrupt_cookie_log_pub_key:
            return

        # .. setting_name: UNUSUAL_COOKIE_HEADER_LOG_CHUNK
        # .. setting_default: 9000
        # .. setting_description: If necessary, logs data in chunks of this size, splitting across
        #   multiple log messages. This should be set with your deployment's maximum log message
        #   size in mind. Setting it too high may result in truncated messages, which will prevent
        #   decryption (CryptoError). Since there is a relatively constant-size prefix on all log
        #   messages (date, module, etc.) this chunk size should be at least 500 bytes shorter than
        #   the max log message size to provide a reasonable margin of safety.
        chunk_size = getattr(settings, 'UNUSUAL_COOKIE_HEADER_LOG_CHUNK', 9000)

        header_data = json.dumps(dict(request.headers.items()))
        enc_output = encrypt_for_log(header_data, corrupt_cookie_log_pub_key)
        msg = f"All headers for request with corrupted cookies (count={corrupt_cookie_count}): {enc_output}"

        for piece in split_ascii_log_message(msg, chunk_size):
            log.info(piece)


# This function should be cleaned up and made into a general logging utility, but it will first
# need some work to make it able to handle multibyte characters.
#
# - If you split on character count, then you may exceed the line size if there are a large number
#   of multibyte characters.
# - If you split on byte count, then multibyte characters like "犬" run a risk of being split
#   between the bytes, resulting in b'ç' and 'b\x8a¬' and no easy way to reconstruct the data.
#
# It may be that there's some library that already does this. In the meantime, this code should be
# sufficient for corrupted cookie logging, with its known-character-range code.
#
# If it is separated out, it should get its own setting, maybe `LONG_LOG_MESSAGE_CHUNK_SIZE`.
def split_ascii_log_message(msg, chunk_size):
    """
    Generator that splits a message string into chunks of at most ``chunk_size`` characters.

    Message must consist of single-byte (ASCII-range) characters, otherwise chunks will
    not reliably be capped at the ``chunk_size``.

    A small collation tag will be added to the end of each chunk, and it may not be possible
    to reliably predict the length of a log message's prefix (in its final output format), so
    the ``chunk_size`` should be set considerably smaller than max log message size.
    """
    chunk_count = math.ceil(len(msg) / chunk_size)
    if chunk_count <= 1:
        yield msg  # no need for continuation messages
    else:
        # Generate a unique-enough collation ID for this message.
        h = hashlib.shake_128(msg.encode()).digest(6)  # pylint/#4039 pylint: disable=too-many-function-args
        group_id = base64.b64encode(h).decode().rstrip('=')

        for i in range(chunk_count):
            chunk = msg[i*chunk_size:(i + 1)*chunk_size]
            if i < chunk_count - 1:
                continue_desc = "continues"
            else:
                continue_desc = "final"
            yield f"{chunk} [chunk #{i + 1}, group={group_id} {continue_desc}]"



"""
Tests for the code_owner monitoring middleware
"""
from unittest import TestCase
from unittest.mock import ANY, MagicMock, Mock, call, patch

import ddt
from django.test import RequestFactory, override_settings
from django.urls import re_path
from django.views.generic import View

from edx_django_utils.monitoring import CodeOwnerMonitoringMiddleware
from edx_django_utils.monitoring.internal.code_owner.utils import clear_cached_mappings

from .mock_views import MockViewTest


class MockMiddlewareViewTest(View):
    pass


urlpatterns = [
    re_path(r'^middleware-test/$', MockMiddlewareViewTest.as_view()),
    re_path(r'^test/$', MockViewTest.as_view()),
]

SET_CUSTOM_ATTRIBUTE_MOCK = MagicMock()


# Enables the same mock to be used from different modules, using
#   patch with new_callable=get_set_custom_attribute_mock
def get_set_custom_attribute_mock():
    return SET_CUSTOM_ATTRIBUTE_MOCK


@ddt.ddt
class CodeOwnerMetricMiddlewareTests(TestCase):
    """
    Tests for the code_owner monitoring utility functions
    """
    urls = 'lms.djangoapps.monitoring.tests.test_middleware.test_urls'

    def setUp(self):
        super().setUp()
        clear_cached_mappings()
        SET_CUSTOM_ATTRIBUTE_MOCK.reset_mock()
        self.mock_get_response = Mock()
        self.middleware = CodeOwnerMonitoringMiddleware(self.mock_get_response)

    def test_init(self):
        self.assertEqual(self.middleware.get_response, self.mock_get_response)

    def test_request_call(self):
        self.mock_get_response.return_value = 'test-response'
        request = Mock()
        self.assertEqual(self.middleware(request), 'test-response')

    _REQUEST_PATH_TO_MODULE_PATH = {
        '/middleware-test/': 'edx_django_utils.monitoring.tests.code_owner.test_middleware',
        '/test/': 'edx_django_utils.monitoring.tests.code_owner.mock_views',
    }

    @override_settings(
        CODE_OWNER_MAPPINGS={'team-red': ['edx_django_utils.monitoring.tests.code_owner.mock_views']},
        CODE_OWNER_THEMES={'team': ['team-red']},
        ROOT_URLCONF=__name__,
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.utils.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.middleware.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @ddt.data(
        ('/middleware-test/', None),
        ('/test/', 'team-red'),
    )
    @ddt.unpack
    def test_code_owner_path_mapping_hits_and_misses(
        self, request_path, expected_owner, mock_set_custom_attribute, _
    ):
        request = RequestFactory().get(request_path)
        self.middleware(request)
        expected_path_module = self._REQUEST_PATH_TO_MODULE_PATH[request_path]
        self._assert_code_owner_custom_attributes(
            mock_set_custom_attribute, expected_code_owner=expected_owner, path_module=expected_path_module,
            check_theme_and_squad=True
        )

        mock_set_custom_attribute.reset_mock()
        self.middleware.process_exception(request, None)
        self._assert_code_owner_custom_attributes(
            mock_set_custom_attribute, expected_code_owner=expected_owner, path_module=expected_path_module,
            check_theme_and_squad=True
        )

    @override_settings(
        CODE_OWNER_MAPPINGS={
            'team-red': ['edx_django_utils.monitoring.tests.code_owner.mock_views'],
            'team-blue': ['*'],
        },
        ROOT_URLCONF=__name__,
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.utils.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.middleware.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @ddt.data(
        ('/middleware-test/', 'team-blue'),
        ('/test/', 'team-red'),
    )
    @ddt.unpack
    def test_code_owner_path_mapping_with_catch_all(
        self, request_path, expected_owner, mock_set_custom_attribute, _
    ):
        request = RequestFactory().get(request_path)
        self.middleware(request)
        expected_path_module = self._REQUEST_PATH_TO_MODULE_PATH[request_path]
        self._assert_code_owner_custom_attributes(
            mock_set_custom_attribute, expected_code_owner=expected_owner, path_module=expected_path_module
        )

    @override_settings(
        CODE_OWNER_MAPPINGS={'team-red': ['edx_django_utils.monitoring.tests.code_owner.mock_views']},
        ROOT_URLCONF=__name__,
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.utils.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.middleware.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @patch('newrelic.agent')
    @ddt.data(
        (
            'edx_django_utils.monitoring.tests.code_owner.test_middleware',
            'edx_django_utils.monitoring.tests.code_owner.test_middleware:MockMiddlewareViewTest',
            None
        ),
        (
            'edx_django_utils.monitoring.tests.code_owner.mock_views',
            'edx_django_utils.monitoring.tests.code_owner.mock_views:MockViewTest',
            'team-red'
        ),
    )
    @ddt.unpack
    def test_code_owner_transaction_mapping_hits_and_misses(
        self, path_module, transaction_name, expected_owner, mock_newrelic_agent, mock_set_custom_attribute, _
    ):
        mock_newrelic_agent.current_transaction().name = transaction_name
        request = RequestFactory().get('/bad/path/')
        self.middleware(request)
        self._assert_code_owner_custom_attributes(
            mock_set_custom_attribute, expected_code_owner=expected_owner, path_module=path_module,
            transaction_name=transaction_name
        )

        mock_set_custom_attribute.reset_mock()
        self.middleware.process_exception(request, None)
        self._assert_code_owner_custom_attributes(
            mock_set_custom_attribute, expected_code_owner=expected_owner, path_module=path_module,
            transaction_name=transaction_name
        )

    @override_settings(
        CODE_OWNER_MAPPINGS={
            'team-red': ['edx_django_utils.monitoring.tests.code_owner.mock_views'],
            'team-blue': ['*'],
        },
        ROOT_URLCONF=__name__,
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.utils.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.middleware.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @patch('newrelic.agent')
    @ddt.data(
        (
            'edx_django_utils.monitoring.tests.code_owner.test_middleware',
            'edx_django_utils.monitoring.tests.code_owner.test_middleware:MockMiddlewareViewTest',
            'team-blue'
        ),
        (
            'edx_django_utils.monitoring.tests.code_owner.mock_views',
            'edx_django_utils.monitoring.tests.code_owner.mock_views:MockViewTest',
            'team-red'
        ),
    )
    @ddt.unpack
    def test_code_owner_transaction_mapping_with_catch_all(
        self, path_module, transaction_name, expected_owner, mock_newrelic_agent, mock_set_custom_attribute, _
    ):
        mock_newrelic_agent.current_transaction().name = transaction_name
        request = RequestFactory().get('/bad/path/')
        self.middleware(request)
        self._assert_code_owner_custom_attributes(
            mock_set_custom_attribute, expected_code_owner=expected_owner, path_module=path_module,
            transaction_name=transaction_name
        )

    @override_settings(
        CODE_OWNER_MAPPINGS={'team-red': ['edx_django_utils.monitoring.tests.code_owner.mock_views']},
        ROOT_URLCONF=__name__,
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.utils.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.middleware.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @patch('newrelic.agent')
    def test_code_owner_transaction_mapping_error(self, mock_newrelic_agent, mock_set_custom_attribute, _):
        mock_newrelic_agent.current_transaction = Mock(side_effect=Exception('forced exception'))
        request = RequestFactory().get('/bad/path/')
        self.middleware(request)
        self._assert_code_owner_custom_attributes(
            mock_set_custom_attribute, has_path_error=True, has_transaction_error=True
        )

    @patch('edx_django_utils.monitoring.internal.code_owner.utils.set_custom_attribute')
    def test_code_owner_no_mappings(self, mock_set_custom_attribute):
        request = RequestFactory().get('/test/')
        self.middleware(request)
        mock_set_custom_attribute.assert_not_called()

    @patch('edx_django_utils.monitoring.internal.code_owner.utils.set_custom_attribute')
    def test_code_owner_transaction_no_mappings(self, mock_set_custom_attribute):
        request = RequestFactory().get('/bad/path/')
        self.middleware(request)
        mock_set_custom_attribute.assert_not_called()

    @override_settings(
        CODE_OWNER_MAPPINGS={'team-red': ['lms.djangoapps.monitoring.tests.mock_views']},
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.utils.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.middleware.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    def test_no_resolver_for_path_and_no_transaction(self, mock_set_custom_attribute, _):
        request = RequestFactory().get('/bad/path/')
        self.middleware(request)
        self._assert_code_owner_custom_attributes(
            mock_set_custom_attribute, has_path_error=True, has_transaction_error=True
        )

    @override_settings(
        CODE_OWNER_MAPPINGS={'team-red': ['*']},
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.utils.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    @patch(
        'edx_django_utils.monitoring.internal.code_owner.middleware.set_custom_attribute',
        new_callable=get_set_custom_attribute_mock
    )
    def test_catch_all_with_errors(self, mock_set_custom_attribute, _):
        request = RequestFactory().get('/bad/path/')
        self.middleware(request)
        self._assert_code_owner_custom_attributes(
            mock_set_custom_attribute, has_path_error=True, has_transaction_error=True, expected_code_owner='team-red'
        )

    @override_settings(
        CODE_OWNER_MAPPINGS=['invalid_setting_as_list'],
        ROOT_URLCONF=__name__,
    )
    def test_load_config_with_invalid_dict(self):
        request = RequestFactory().get('/test/')
        with self.assertRaises(TypeError):
            self.middleware(request)

    def _assert_code_owner_custom_attributes(self, mock_set_custom_attribute, expected_code_owner=None,
                                             path_module=None, has_path_error=False,
                                             transaction_name=None, has_transaction_error=False,
                                             check_theme_and_squad=False):
        """ Performs a set of assertions around having set the proper custom attributes. """
        call_list = []
        if expected_code_owner:
            call_list.append(call('code_owner', expected_code_owner))
            if check_theme_and_squad:
                call_list.append(call('code_owner_theme', expected_code_owner.split('-')[0]))
                call_list.append(call('code_owner_squad', expected_code_owner.split('-')[1]))
        if path_module:
            call_list.append(call('code_owner_module', path_module))
        if has_path_error:
            call_list.append(call('code_owner_path_error', ANY))
        if transaction_name:
            call_list.append(call('code_owner_transaction_name', transaction_name))
        if has_transaction_error:
            call_list.append(call('code_owner_transaction_error', ANY))
        # TODO: Remove this list filtering once the ``deprecated_broad_except_XXX`` custom attributes have been removed.
        actual_filtered_call_list = [
            mock_call
            for mock_call in mock_set_custom_attribute.call_args_list
            if not mock_call[0][0].startswith('deprecated_')
        ]
        mock_set_custom_attribute.assert_has_calls(call_list, any_order=True)
        self.assertEqual(
            len(actual_filtered_call_list), len(call_list),
            f'Expected calls {call_list} vs actual calls {actual_filtered_call_list}'
        )



"""
Tests for Content-Security-Policy middleware.
"""

from unittest import TestCase
from unittest.mock import Mock, patch

import ddt
import pytest
from django.core.exceptions import MiddlewareNotUsed
from django.test import override_settings

import edx_django_utils.security.csp.middleware as csp


@ddt.ddt
class TestLoadHeaders(TestCase):
    """Test loading of headers from settings."""

    @ddt.unpack
    @ddt.data(
        # Empty settings
        [{}, {}],
        # The reporting URL and endpoint names alone don't do anything
        [{"CSP_STATIC_REPORTING_URI": "http://localhost"}, {}],
        [{"CSP_STATIC_REPORTING_NAME": "default"}, {}],
        [
            {
                "CSP_STATIC_REPORTING_URI": "http://localhost",
                "CSP_STATIC_REPORTING_NAME": "default"
            },
            {},
        ],
        # Just the enforcement header
        [
            {"CSP_STATIC_ENFORCE": "default-src https:"},
            {'Content-Security-Policy': "default-src https:"},
        ],
        # Just the reporting header
        [
            {"CSP_STATIC_REPORT_ONLY": "default-src 'none'"},
            {'Content-Security-Policy-Report-Only': "default-src 'none'"},
        ],
        # Reporting URL is automatically appended to headers
        [
            {
                "CSP_STATIC_ENFORCE": "default-src https:",
                "CSP_STATIC_REPORT_ONLY": "default-src 'none'",
                "CSP_STATIC_REPORTING_URI": "http://localhost",
            },
            {
                'Content-Security-Policy': "default-src https:; report-uri http://localhost",
                'Content-Security-Policy-Report-Only': "default-src 'none'; report-uri http://localhost",
            },
        ],
        # ...and when an endpoint name is supplied,
        # Reporting-Endpoints is added and a report-to directive is
        # included.
        [
            {
                "CSP_STATIC_ENFORCE": "default-src https:",
                "CSP_STATIC_REPORT_ONLY": "default-src 'none'",
                "CSP_STATIC_REPORTING_URI": "http://localhost",
                "CSP_STATIC_REPORTING_NAME": "default",
            },
            {
                'Reporting-Endpoints': 'default="http://localhost"',
                'Content-Security-Policy': "default-src https:; report-uri http://localhost; report-to default",
                'Content-Security-Policy-Report-Only': (
                    "default-src 'none'; report-uri http://localhost; report-to default"
                ),
            },
        ],
        # Adding a reporting endpoint name without a URL doesn't change anything.
        [
            {
                "CSP_STATIC_REPORT_ONLY": "default-src 'none'",
                "CSP_STATIC_REPORTING_NAME": "default",
            },
            {'Content-Security-Policy-Report-Only': "default-src 'none'"},
        ],
        # Any newlines and trailing semicolon are stripped.
        [
            {
                "CSP_STATIC_REPORT_ONLY": "default-src 'self';   \n \t  frame-src 'none';  \n ",
                "CSP_STATIC_REPORTING_URI": "http://localhost",
            },
            {
                'Content-Security-Policy-Report-Only': (
                    "default-src 'self'; frame-src 'none'; "
                    "report-uri http://localhost"
                ),
            },
        ],
    )
    def test_load_headers(self, settings, headers):
        with override_settings(**settings):
            assert csp._load_headers() == headers  # pylint: disable=protected-access


@ddt.ddt
class TestHeaderManipulation(TestCase):
    """Test _append_headers"""

    @ddt.unpack
    @ddt.data(
        [{}, {}, {}],
        [
            {'existing': 'aaa', 'multi': '111'},
            {'multi': '222', 'new': 'xxx'},
            {'existing': 'aaa', 'multi': '111, 222', 'new': 'xxx'},
        ],
    )
    def test_append_headers(self, response_headers, more_headers, expected):
        csp._append_headers(response_headers, more_headers)  # pylint: disable=protected-access
        assert response_headers == expected


@ddt.ddt
class TestCSPMiddleware(TestCase):
    """Test the actual middleware."""

    def setUp(self):
        super().setUp()
        self.fake_response = Mock()
        self.fake_response.headers = {'Existing': 'something'}

    def test_make_middleware_unused(self):
        with pytest.raises(MiddlewareNotUsed):
            csp.content_security_policy_middleware(lambda _: self.fake_response)

    @override_settings(CSP_STATIC_ENFORCE="default-src: https:")
    def test_make_middleware_configured(self):
        handler = csp.content_security_policy_middleware(lambda _: self.fake_response)

        assert handler(Mock()) is self.fake_response

        # Headers have been mutated in place (if flag enabled)
        assert self.fake_response.headers == {
            'Existing': 'something',
            'Content-Security-Policy': 'default-src: https:',
        }



"""
Cache utilities.
"""
import hashlib
import threading

from django.core.cache import cache as django_cache
from django.core.cache.backends.base import DEFAULT_TIMEOUT
from django.utils.encoding import force_str

FORCE_CACHE_MISS_PARAM = 'force_cache_miss'
DEFAULT_NAMESPACE = 'edx_django_utils.cache'
DEFAULT_REQUEST_CACHE_NAMESPACE = f'{DEFAULT_NAMESPACE}.default'
SHOULD_FORCE_CACHE_MISS_KEY = 'edx_django_utils.cache.should_force_cache_miss'

_CACHE_MISS = object()


def get_cache_key(**kwargs):
    """
    Get MD5 encoded cache key for given arguments.

    **Note:** We convert keyword arguments to their string form to build the cache key. So do not pass
    arguments that can't be converted to strings.  Also, don't pass arguments that may not consistently
    be converted to the same string, like an unsorted dict.
    Here is the format of key before MD5 encryption.
        key1:value1__key2:value2 ...
    Example:
        >>> get_cache_key(site_domain="example.com", resource="catalogs")
        # Here is key format for above call
        # "site_domain:example.com__resource:catalogs"
        a54349175618ff1659dee0978e3149ca
    Arguments:
        **kwargs: Key word arguments that need to be present in cache key.
    Returns:
         An MD5 encoded key uniquely identified by the key word arguments.
    """
    key = '__'.join(['{}:{}'.format(k, force_str(v)) for k, v in sorted(kwargs.items())])

    return hashlib.md5(key.encode('utf-8')).hexdigest()


class _RequestCache(threading.local):
    """
    A thread-local for storing the per-request caches.

    The data is a dict of dicts, keyed by namespace.
    """
    def __init__(self):
        super().__init__()
        self._data = {}

    def clear(self):
        """
        Clears all data for all namespaces.
        """
        self._data = {}

    def data(self, namespace):
        """
        Gets the thread.local data (dict) for a given namespace.

        Args:
            namespace (string): The namespace, or key, of the data dict.

        Returns:
            (dict)

        """
        assert namespace

        if namespace in self._data:
            return self._data[namespace]

        new_data = {}
        self._data[namespace] = new_data
        return new_data


# Singleton with thread-local cache
_REQUEST_CACHE = _RequestCache()


class RequestCache:
    """
    A namespaced request cache for caching per-request data.
    """

    def __init__(self, namespace=None):
        """
        Creates a request cache with the provided namespace.

        Args:
            namespace (string): (optional) uses 'default' if not provided.
        """
        assert namespace != DEFAULT_REQUEST_CACHE_NAMESPACE,\
            f'Optional namespace can not be {DEFAULT_REQUEST_CACHE_NAMESPACE}.'
        self.namespace = namespace or DEFAULT_REQUEST_CACHE_NAMESPACE

    @classmethod
    def clear_all_namespaces(cls):
        """
        Clears the data for all namespaces.
        """
        _REQUEST_CACHE.clear()

    @property
    def data(self):
        """
        Returns the namespaced cached key/value pairs as a dict.
        """
        return _REQUEST_CACHE.data(self.namespace)

    def clear(self):
        """
        Clears data for the namespaced request cache.
        """
        self.data.clear()

    def get_cached_response(self, key):
        """
        Retrieves a CachedResponse for the provided key.

        Args:
            key (string)

        Returns:
            A CachedResponse with is_found status and value.

        """
        cached_value = self.data.get(key, _CACHE_MISS)
        is_found = cached_value is not _CACHE_MISS
        return CachedResponse(is_found, key, cached_value)

    def set(self, key, value):
        """
        Caches the value for the provided key.

        Args:
            key (string)
            value (object)

        """
        self.data[key] = value

    def setdefault(self, key, value):
        """
        Sets the value for the provided key if it has not yet been set.

        Args:
            key (string)
            value (object)

        """
        self.data.setdefault(key, value)

    def delete(self, key):
        """
        Deletes the cached value for the provided key.

        Args:
            key (string)

        """
        if key in self.data:
            del self.data[key]


DEFAULT_REQUEST_CACHE = RequestCache()


class TieredCache:
    """
    A two tiered caching object with a request cache backed by a django cache.
    """

    @classmethod
    def get_cached_response(cls, key):
        """
        Retrieves a CachedResponse for the provided key.

        Args:
            key (string)

        Returns:
            A CachedResponse with is_found status and value.

        """
        request_cached_response = DEFAULT_REQUEST_CACHE.get_cached_response(key)
        if not request_cached_response.is_found:
            django_cached_response = cls._get_cached_response_from_django_cache(key)
            cls._set_request_cache_if_django_cache_hit(key, django_cached_response)
            return django_cached_response

        return request_cached_response

    @staticmethod
    def set_all_tiers(key, value, django_cache_timeout=DEFAULT_TIMEOUT):
        """
        Caches the value for the provided key in both the request cache and the
        django cache.

        Args:
            key (string)
            value (object)
            django_cache_timeout (int): (Optional) Timeout used to determine
                if and for how long to cache in the django cache. A timeout of
                0 will skip the django cache. If timeout is provided, use that
                timeout for the key; otherwise use the default cache timeout.

        """
        DEFAULT_REQUEST_CACHE.set(key, value)
        django_cache.set(key, value, django_cache_timeout)

    @staticmethod
    def delete_all_tiers(key):
        """
        Deletes the cached value for the provided key in both the request cache and the
        django cache.

        Args:
            key (string)

        """
        DEFAULT_REQUEST_CACHE.delete(key)
        django_cache.delete(key)

    @staticmethod
    def dangerous_clear_all_tiers():
        """
        This clears both the default request cache and the entire django
        backing cache.

        Important: This should probably only be called for testing purposes.

        TODO: Move CacheIsolationMixin from edx-platform to edx-django-utils
        and kill this method.

        """
        DEFAULT_REQUEST_CACHE.clear()
        django_cache.clear()

    @staticmethod
    def _get_cached_response_from_django_cache(key):
        """
        Retrieves a CachedResponse for the given key from the django cache.

        If the request was set to force cache misses, then this will always
        return a cache miss response.

        Args:
            key (string)

        Returns:
            A CachedResponse with is_found status and value.

        """
        if TieredCache._should_force_django_cache_miss():
            return CachedResponse(is_found=False, key=key, value=None)

        cached_value = django_cache.get(key, _CACHE_MISS)
        is_found = cached_value is not _CACHE_MISS
        return CachedResponse(is_found, key, cached_value)

    @staticmethod
    def _set_request_cache_if_django_cache_hit(key, django_cached_response):
        """
        Sets the value in the request cache if the django cached response was a hit.

        Args:
            key (string)
            django_cached_response (CachedResponse)

        """
        if django_cached_response.is_found:
            DEFAULT_REQUEST_CACHE.set(key, django_cached_response.value)

    @staticmethod
    def _get_and_set_force_cache_miss(request):
        """
        Gets value for request query parameter FORCE_CACHE_MISS
        and sets it in the default request cache.

        This functionality is only available for staff.

        Example:
            http://clobert.com/api/v1/resource?force_cache_miss=true

        """
        if not (request.user and request.user.is_active and request.user.is_staff):
            force_cache_miss = False
        else:
            force_cache_miss = request.GET.get(FORCE_CACHE_MISS_PARAM, 'false').lower() == 'true'
        DEFAULT_REQUEST_CACHE.set(SHOULD_FORCE_CACHE_MISS_KEY, force_cache_miss)

    @classmethod
    def _should_force_django_cache_miss(cls):
        """
        Returns True if the tiered cache should force a cache miss for the
        django cache, and False otherwise.

        """
        cached_response = DEFAULT_REQUEST_CACHE.get_cached_response(SHOULD_FORCE_CACHE_MISS_KEY)
        return False if not cached_response.is_found else cached_response.value


class CachedResponseError(Exception):
    """
    Error used when CachedResponse is misused.
    """
    USAGE_MESSAGE = 'CachedResponse was misused. Try the attributes is_found, value or key.'

    def __init__(self, message=USAGE_MESSAGE):  # pylint: disable=useless-super-delegation
        super().__init__(message)


class CachedResponse:
    """
    Represents a cache response including is_found status and value.
    """
    def __init__(self, is_found, key, value):
        """
        Creates a cached response object.

        Args:
            is_found (bool): True if the key was found in the cache, False
                otherwise.
            key (string): The key originally used to retrieve the value.
            value (object)
        """
        self.key = key
        self.is_found = is_found
        if self.is_found:
            self.value = value

    def __repr__(self):
        # Important: Do not include the cached value to help avoid any security
        # leaks that could happen if these are logged.
        return f'''CachedResponse(is_found={self.is_found}, key={self.key}, value='*****')'''

    def get_value_or_default(self, default):
        """
        Returns value for a cache hit, or the passed default for a cache miss.

        This method is safe to use, even for a cache_miss.

        WARNING: Never pass None as the default and then test the return value
        of this method. Use is_found instead for any checks.

        """
        return default if not self.is_found else self.value

    def __bool__(self):
        raise CachedResponseError()

    def __eq__(self, other):
        if not isinstance(other, CachedResponse):
            raise CachedResponseError()
        if self.is_found != other.is_found:
            return False

        if self.is_found:
            return (self.key, self.value) == (other.key, other.value)
        else:
            return self.key == other.key  # cache misses have no value attribute

    def __ne__(self, other):
        """Overrides the default implementation (unnecessary in Python 3)"""
        return not self.__eq__(other)
